"""This module defines the ChEMBL ETL methods."""
import logging
import os
import shutil
import sqlite3
from typing import Optional, List, Dict

import chembl_downloader
import bioversions

from therapy.etl.base import DiseaseIndicationBase
from therapy.schemas import SourceName, NamespacePrefix, ApprovalRating, \
    SourceMeta


logger = logging.getLogger("therapy")
logger.setLevel(logging.DEBUG)


class ChEMBL(DiseaseIndicationBase):
    """Class for ChEMBL ETL methods."""

    def _download_data(self) -> None:
        """Download latest ChEMBL database file from EBI."""
        logger.info("Retrieving source data for ChEMBL")
        os.environ["PYSTOW_HOME"] = str(self._src_dir.parent.absolute())
        tmp_path = chembl_downloader.download_extract_sqlite()
        shutil.move(tmp_path, self._src_dir)
        shutil.rmtree(tmp_path.parent.parent.parent)
        logger.info("Successfully retrieved source data for ChEMBL")

    def _extract_data(self, use_existing: bool = False) -> None:
        """Extract data from the ChEMBL source.

        :param bool use_existing: if True, don't try to fetch latest source data
        """
        super()._extract_data(use_existing)
        conn = sqlite3.connect(self._src_file)
        conn.row_factory = sqlite3.Row
        self._conn = conn
        self._cursor = conn.cursor()

    @staticmethod
    def _unwrap_group_concat(value: Optional[str]) -> List[str]:
        """Process concatenated values retrieved from ChEMBL DB.
        :param Optional[str] value: the value provided by the cursor
        :return: List of separated values (empty if none present)
        """
        if value:
            return value.split("||")
        else:
            return []

    @staticmethod
    def _get_approval_rating(value: int) -> ApprovalRating:
        """Standardize approval rating value
        :param int value: value retrieved from ChEMBL database
        :return: instantiated ApprovalRating
        :raise: ValueError if invalid value is provided
        """
        if value == 0:
            return ApprovalRating.CHEMBL_0
        elif value == 1:
            return ApprovalRating.CHEMBL_1
        elif value == 2:
            return ApprovalRating.CHEMBL_2
        elif value == 3:
            return ApprovalRating.CHEMBL_3
        elif value == 4:
            return ApprovalRating.CHEMBL_4
        else:
            raise ValueError(f"Unrecognized approval rating: {value}")

    def _get_indications(self, value: Optional[str]) -> List[Dict]:
        """Construct indication objects to prepare for loading into DB
        :param Optional[str] value: `indication` value fetched from ChEMBL DB
        :return: List containing indication dicts
        """
        if value:
            indications = []
            indication_groups = value.split("|||")
            for group in set(indication_groups):
                ind_group = group.split("||")
                phase = self._get_approval_rating(int(ind_group[4]))
                indication = {}
                for i, term in enumerate(ind_group[:4]):
                    normalized_disease_id = self._normalize_disease(term)
                    if normalized_disease_id is not None:
                        label = ind_group[2] if i % 2 == 0 else ind_group[3]
                        disease_id = ind_group[0] if i % 2 == 0 else ind_group[1]
                        indication = {
                            "disease_id": disease_id,
                            "disease_label": label,
                            "normalized_disease_id": normalized_disease_id,
                            "supplemental_info": {"chembl_max_phase_for_ind": phase}
                        }
                        break
                if not indication:
                    indication = {
                        "disease_id": ind_group[0],
                        "disease_label": ind_group[2],
                        "supplemental_info": {"chembl_max_phase_for_ind": phase}
                    }
                indications.append(indication)
            return indications
        else:
            return []

    def _transform_data(self) -> None:
        """Transform SQLite data and load to DB."""
        query = """
        SELECT
            md.chembl_id,
            md.molregno,
            md.pref_name,
            md.max_phase,
            md.withdrawn_flag,
            ms.aliases,
            tn.trade_names,
            d.indications
        FROM molecule_dictionary md
        LEFT JOIN (
            SELECT molregno, GROUP_CONCAT(synonyms, '||') aliases
            FROM (
                SELECT DISTINCT molregno, synonyms
                FROM molecule_synonyms
                WHERE syn_type != 'TRADE_NAME'
            ) GROUP BY molregno
        ) ms on md.molregno = ms.molregno
        LEFT JOIN (
            SELECT molregno, GROUP_CONCAT(trade_name, '||') trade_names
            FROM (
                SELECT DISTINCT molregno, trade_name
                FROM (
                    SELECT molregno, synonyms as trade_name
                    FROM molecule_synonyms
                    WHERE syn_type == 'TRADE_NAME'
                    UNION
                    SELECT molregno, trade_name
                    FROM (
                        SELECT f.molregno, p.trade_name
                        FROM formulations f
                        LEFT JOIN products p
                        ON f.product_id = p.product_id
                    )
                )
            ) GROUP BY molregno
        ) tn on md.molregno = tn.molregno
        LEFT JOIN (
            SELECT molregno, GROUP_CONCAT(indication, '|||') as indications
            FROM (
                SELECT DISTINCT d.molregno,
                'mesh:' || d.mesh_id || '||' || d.efo_id || '||' || d.mesh_heading
                 || '||' || d.efo_term || '||' || d.max_phase_for_ind as indication
                FROM drug_indication AS d
                WHERE (
                    d.molregno IS NOT NULL
                    AND d.mesh_heading IS NOT NULL
                    AND d.mesh_id IS NOT NULL
                    AND d.efo_term IS NOT NULL
                    AND d.efo_id IS NOT NULL
                    AND d.max_phase_for_ind IS NOT NULL
                )
            ) GROUP BY molregno
        ) d on md.molregno = d.molregno
        GROUP BY md.molregno;
        """
        self._cursor.execute(query)

        for row in self._cursor:
            appr_ratings = []
            max_phase = self._get_approval_rating(row["max_phase"])
            if max_phase is not None:
                appr_ratings.append(max_phase)
            if row["withdrawn_flag"] == 1:
                appr_ratings.append(ApprovalRating.CHEMBL_WITHDRAWN)

            has_indication = self._get_indications(row["indications"])

            params = {
                "concept_id": f"{NamespacePrefix.CHEMBL.value}:{row['chembl_id']}",
                "label": row["pref_name"],
                "approval_ratings": appr_ratings,
                "aliases": self._unwrap_group_concat(row["aliases"]),
                "trade_names": self._unwrap_group_concat(row["trade_names"]),
                "has_indication": has_indication
            }
            self._load_therapy(params)
        self._conn.close()

    def _load_meta(self) -> None:
        """Add ChEMBL metadata."""
        metadata = SourceMeta(data_license="CC BY-SA 3.0",
                              data_license_url="https://creativecommons.org/licenses/by-sa/3.0/",  # noqa: E501
                              version=self._version,
                              data_url=bioversions.resolve("chembl").homepage,
                              rdp_url="http://reusabledata.org/chembl.html",
                              data_license_attributes={
                                  "non_commercial": False,
                                  "share_alike": True,
                                  "attribution": True
                              })
        params = dict(metadata)
        params["src_name"] = SourceName.CHEMBL.value
        self.database.metadata.put_item(Item=params)



"""A base class for extraction, transformation, and loading of data."""
from abc import ABC, abstractmethod
import ftplib
from pathlib import Path
import logging
from typing import List, Dict, Optional, Callable
import os
import zipfile
import tempfile
import re
import json
from functools import lru_cache

from pydantic import ValidationError
import requests
import bioversions
from disease.query import QueryHandler as DiseaseNormalizer

from therapy import APP_ROOT, ITEM_TYPES, DownloadException
from therapy.schemas import Drug, SourceName
from therapy.database import Database
from therapy.etl.rules import Rules


logger = logging.getLogger("therapy")
logger.setLevel(logging.DEBUG)

DEFAULT_DATA_PATH: Path = APP_ROOT / "data"


class Base(ABC):
    """The ETL base class.

    Default methods are declared to provide basic functions for core source
    data-gathering phases (extraction, transformation, loading), as well
    as some common subtasks (getting most recent version, downloading data
    from an FTP server). Classes should expand or reimplement these methods as
    needed.
    """

    def __init__(self, database: Database, data_path: Path = DEFAULT_DATA_PATH) -> None:
        """Extract from sources.

        :param Database database: application database object
        :param Path data_path: path to app data directory
        """
        self._name = self.__class__.__name__
        self.database = database
        self._src_dir: Path = Path(data_path / self._name.lower())
        self._added_ids: List[str] = []
        self._rules = Rules(SourceName(self._name))

    def perform_etl(self, use_existing: bool = False) -> List[str]:
        """Public-facing method to begin ETL procedures on given data.
        Returned concept IDs can be passed to Merge method for computing
        merged concepts.

        :param bool use_existing: if True, don't try to retrieve latest source data
        :return: list of concept IDs which were successfully processed and
            uploaded.
        """
        self._extract_data(use_existing)
        self._load_meta()
        self._transform_data()
        return self._added_ids

    def get_latest_version(self) -> str:
        """Get most recent version of source data. Should be overriden by
        sources not added to Bioversions yet, or other special-case sources.
        :return: most recent version, as a str
        """
        return bioversions.get_version(self.__class__.__name__)

    @abstractmethod
    def _download_data(self) -> None:
        """Acquire source data and deposit in a usable form with correct file
        naming conventions (generally, `<source>_<version>.<filetype>`, or
        `<source>_<subset>_<version>.<filetype>` if sources require multiple
        files). Shouldn't set any instance attributes.
        """
        raise NotImplementedError

    def _zip_handler(self, dl_path: Path, outfile_path: Path) -> None:
        """Provide simple callback function to extract the largest file within a given
        zipfile and save it within the appropriate data directory.
        :param Path dl_path: path to temp data file
        :param Path outfile_path: path to save file within
        """
        with zipfile.ZipFile(dl_path, "r") as zip_ref:
            if len(zip_ref.filelist) > 1:
                files = sorted(
                    zip_ref.filelist, key=lambda z: z.file_size, reverse=True
                )
                target = files[0]
            else:
                target = zip_ref.filelist[0]
            target.filename = outfile_path.name
            zip_ref.extract(target, path=outfile_path.parent)
        os.remove(dl_path)

    @staticmethod
    def _http_download(
        url: str,
        outfile_path: Path,
        headers: Optional[Dict] = None,
        handler: Optional[Callable[[Path, Path], None]] = None,
    ) -> None:
        """Perform HTTP download of remote data file.
        :param str url: URL to retrieve file from
        :param Path outfile_path: path to where file should be saved. Must be an actual
            Path instance rather than merely a pathlike string.
        :param Optional[Dict] headers: Any needed HTTP headers to include in request
        :param Optional[Callable[[Path, Path], None]] handler: provide if downloaded
            file requires additional action, e.g. it's a zip file.
        """
        if handler:
            dl_path = Path(tempfile.gettempdir()) / "therapy_dl_tmp"
        else:
            dl_path = outfile_path
        # use stream to avoid saving download completely to memory
        with requests.get(url, stream=True, headers=headers) as r:
            try:
                r.raise_for_status()
            except requests.HTTPError:
                raise DownloadException(
                    f"Failed to download {outfile_path.name} from {url}."
                )
            with open(dl_path, "wb") as h:
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:
                        h.write(chunk)
        if handler:
            handler(dl_path, outfile_path)

    def _ftp_download(self, host: str, host_dir: str, host_fn: str) -> None:
        """Download data file from FTP site.
        :param str host: Source's FTP host name
        :param str host_dir: Data directory located on FTP site
        :param str host_fn: Filename on FTP site to be downloaded
        """
        try:
            with ftplib.FTP(host) as ftp:
                ftp.login()
                logger.debug(f"FTP login to {host} was successful")
                ftp.cwd(host_dir)
                with open(self._src_dir / host_fn, "wb") as fp:
                    ftp.retrbinary(f"RETR {host_fn}", fp.write)
        except ftplib.all_errors as e:
            logger.error(f"FTP download failed: {e}")
            raise Exception(e)

    def _parse_version(
        self, file_path: Path, pattern: Optional[re.Pattern] = None
    ) -> str:
        """Get version number from provided file path.

        :param Path file_path: path to located source data file
        :param Optional[re.Pattern] pattern: regex pattern to use
        :return: source data version
        :raises: FileNotFoundError if version parsing fails
        """
        if pattern is None:
            pattern = re.compile(type(self).__name__.lower() + r"_(.+)\..+")
        matches = re.match(pattern, file_path.name)
        if matches is None:
            raise FileNotFoundError
        else:
            return matches.groups()[0]

    def _get_existing_files(self) -> List[Path]:
        """Get existing source files from data directory.
        :return: sorted list of file objects
        """
        return list(sorted(self._src_dir.glob(f"{self._name.lower()}_*.*")))

    def _extract_data(self, use_existing: bool = False) -> None:
        """Get source file from data directory.

        This method should ensure the source data directory exists, acquire source data,
        set the source version value, and assign the source file location to
        `self._src_file`. Child classes needing additional functionality (like setting
        up a DB cursor, or managing multiple source files) will need to reimplement
        this method. If `use_existing` is True, the version number will be parsed from
        the existing filename; otherwise, it will be retrieved from the data source,
        and if the local file is out-of-date, the newest version will be acquired.

        :param bool use_existing: if True, don't try to fetch latest source data
        """
        self._src_dir.mkdir(exist_ok=True, parents=True)
        src_name = type(self).__name__.lower()
        if use_existing:
            files = self._get_existing_files()
            if len(files) < 1:
                raise FileNotFoundError(f"No source data found for {src_name}")
            self._src_file: Path = files[-1]
            try:
                self._version = self._parse_version(self._src_file)
            except FileNotFoundError:
                raise FileNotFoundError(
                    f"Unable to parse version value from {src_name} source data file "
                    f"located at {self._src_file.absolute().as_uri()} -- "
                    "check filename against schema defined in README: "
                    "https://github.com/cancervariants/therapy-normalization#update-sources"  # noqa: E501
                )
        else:
            self._version = self.get_latest_version()
            fglob = f"{src_name}_{self._version}.*"
            latest = list(self._src_dir.glob(fglob))
            if not latest:
                self._download_data()
                latest = list(self._src_dir.glob(fglob))
            assert len(latest) != 0  # probably unnecessary, but just to be safe
            self._src_file = latest[0]

    @abstractmethod
    def _load_meta(self) -> None:
        """Load source metadata entry."""
        raise NotImplementedError

    @abstractmethod
    def _transform_data(self) -> None:
        """Prepare source data for loading into DB. Individually extract each
        record and call the Base class's `_load_therapy()` method.
        """
        raise NotImplementedError

    def _load_therapy(self, therapy: Dict) -> None:
        """Load individual therapy record into database.
        Additionally, this method takes responsibility for:
            * validating record structure correctness
            * removing duplicates from list-like fields
            * removing empty fields

        :param Dict therapy: valid therapy object.
        """
        therapy = self._rules.apply_rules_to_therapy(therapy)
        try:
            Drug(**therapy)
        except ValidationError as e:
            logger.error(f"Attempted to load invalid therapy: {therapy}")
            raise e

        concept_id = therapy["concept_id"]

        for attr_type, item_type in ITEM_TYPES.items():
            if attr_type in therapy:
                value = therapy[attr_type]
                if value is None or value == []:
                    del therapy[attr_type]
                    continue

                if attr_type == "label":
                    value = value.strip()
                    therapy["label"] = value
                    self.database.add_ref_record(value.lower(), concept_id, item_type)
                    continue

                value_set = {v.strip() for v in value}

                # clean up listlike symbol fields
                if attr_type == "aliases" and "trade_names" in therapy:
                    value = list(value_set - set(therapy["trade_names"]))
                else:
                    value = list(value_set)

                if attr_type in ("aliases", "trade_names"):
                    if "label" in therapy:
                        try:
                            value.remove(therapy["label"])
                        except ValueError:
                            pass

                if len(value) > 20:
                    logger.debug(f"{concept_id} has > 20 {attr_type}.")
                    del therapy[attr_type]
                    continue

                for item in {item.lower() for item in value}:
                    self.database.add_ref_record(item, concept_id, item_type)
                therapy[attr_type] = value

        # compress has_indication
        indications = therapy.get("has_indication")
        if indications:
            therapy["has_indication"] = list(
                {
                    json.dumps(
                        [
                            ind["disease_id"],
                            ind["disease_label"],
                            ind.get("normalized_disease_id"),
                            ind.get("supplemental_info"),
                        ]
                    )
                    for ind in indications
                }
            )
        elif "has_indication" in therapy:
            del therapy["has_indication"]

        # handle detail fields
        approval_attrs = ("approval_ratings", "approval_year")
        for field in approval_attrs:
            if approval_attrs in therapy and therapy[field] is None:
                del therapy[field]

        self.database.add_record(therapy)
        self._added_ids.append(concept_id)


class DiseaseIndicationBase(Base):
    """Base class for sources that require disease normalization capabilities."""

    def __init__(self, database: Database, data_path: Path = DEFAULT_DATA_PATH):
        """Initialize source ETL instance.

        :param therapy.database.Database database: application database
        :param Path data_path: path to normalizer data directory
        """
        super().__init__(database, data_path)
        self.disease_normalizer = DiseaseNormalizer(self.database.endpoint_url)

    @lru_cache(maxsize=64)
    def _normalize_disease(self, query: str) -> Optional[str]:
        """Attempt normalization of disease term.
        :param str query: term to normalize
        :return: ID if successful, None otherwise
        """
        response = self.disease_normalizer.normalize(query)
        if response.match_type > 0:
            return response.disease_descriptor.disease
        else:
            logger.warning(f"Failed to normalize disease term: {query}")
            return None


class SourceFormatException(Exception):
    """Raise when source data formatting is incompatible with the source transformation
    methods: for example, if columns in a CSV file have changed.
    """

    pass



"""Provide ETL methods for HemOnc.org data."""
import logging
from pathlib import Path
from typing import Dict, Tuple, Optional
import csv
import os
import zipfile
import re

import requests
import isodate

from therapy import DownloadException
from therapy.schemas import NamespacePrefix, SourceMeta, SourceName, RecordParams, \
    ApprovalRating
from therapy.etl.base import DiseaseIndicationBase


logger = logging.getLogger("therapy")
logger.setLevel(logging.DEBUG)


class HemOnc(DiseaseIndicationBase):
    """Class for HemOnc.org ETL methods."""

    def get_latest_version(self) -> str:
        """Retrieve latest version of source data.
        :raise: Exception if retrieval is unsuccessful
        """
        response = requests.get("https://dataverse.harvard.edu/api/datasets/export?persistentId=doi:10.7910/DVN/9CY9C6&exporter=dataverse_json")  # noqa: E501
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            logger.error("Unable to retrieve HemOnc version from Harvard Dataverse")
            raise e
        iso_datetime = isodate.parse_datetime(
            response.json()["datasetVersion"]["releaseTime"]
        )
        return iso_datetime.strftime(isodate.isostrf.DATE_EXT_COMPLETE)

    def _zip_handler(self, dl_path: Path, outfile_path: Path) -> None:
        """Extract concepts, rels, and synonyms files from tmp zip file and save to
        data directory.
        :param Path dl_path: path to temp data zipfile
        :param Path outfile_path: directory to save data within
        """
        file_terms = ("concepts", "rels", "synonyms")
        with zipfile.ZipFile(dl_path, "r") as zip_ref:
            for file in zip_ref.filelist:
                for term in file_terms:
                    if term in file.filename:
                        file.filename = f"hemonc_{term}_{self._version}.csv"
                        zip_ref.extract(file, outfile_path)

        os.remove(dl_path)

    def _download_data(self) -> None:
        """Download HemOnc.org source data. Requires Harvard Dataverse API key to be set
        as environment variable DATAVERSE_API_KEY. Instructions for generating an API
        key are available here: https://guides.dataverse.org/en/latest/user/account.html

        :raises: DownloadException if API key environment variable isn't set
        """
        api_key = os.environ.get("DATAVERSE_API_KEY")
        if api_key is None:
            raise DownloadException(
                "Must provide Harvard Dataverse API key in environment variable "
                "DATAVERSE_API_KEY. See "
                "https://guides.dataverse.org/en/latest/user/account.html"
            )
        url = "https://dataverse.harvard.edu//api/access/dataset/:persistentId/?persistentId=doi:10.7910/DVN/9CY9C6"  # noqa: E501
        headers = {"X-Dataverse-key": api_key}
        self._http_download(url, self._src_dir, headers, self._zip_handler)

    def _extract_data(self, use_existing: bool = False) -> None:
        """Get source files from data directory.

        The following files are necessary for data processing:
            hemonc_concepts_<version>.csv
            hemonc_rels_<version>.csv
            hemonc_synonyms_<version>.csv
        This method will attempt to retrieve their latest versions if they are
        unavailable locally.

        :param bool use_existing: if True, don't try to fetch latest source data
        """
        self._src_dir.mkdir(exist_ok=True, parents=True)

        if use_existing:
            concepts = list(sorted(self._src_dir.glob("hemonc_concepts_*.csv")))
            if len(concepts) < 1:
                raise FileNotFoundError("No HemOnc concepts file found")

            src_files: Optional[Tuple] = None
            for concepts_file in concepts[::-1]:
                try:
                    version = self._parse_version(
                        concepts_file,
                        re.compile(r"hemonc_concepts_(.+)\.csv")
                    )
                except FileNotFoundError:
                    raise FileNotFoundError(
                        f"Unable to parse HemOnc version value from concepts file "
                        f"located at {concepts_file.absolute().as_uri()} -- check "
                        "filename against schema defined in README: "
                        "https://github.com/cancervariants/therapy-normalization#update-sources"  # noqa: E501
                    )
                other_files = (
                    self._src_dir / f"hemonc_rels_{version}.csv",
                    self._src_dir / f"hemonc_synonyms_{version}.csv"
                )
                if other_files[0].exists() and other_files[1].exists():
                    self._version = version
                    src_files = (
                        concepts_file,
                        other_files[0],
                        other_files[1]
                    )
                    break
            if src_files is None:
                raise FileNotFoundError(
                    "Unable to find complete HemOnc data set with matching version "
                    "values. Check filenames against schema defined in README: "
                    "https://github.com/cancervariants/therapy-normalization#update-sources"  # noqa: E501
                )
            else:
                self._src_files = src_files
        else:
            self._version = self.get_latest_version()
            data_filenames = (
                self._src_dir / f"hemonc_concepts_{self._version}.csv",
                self._src_dir / f"hemonc_rels_{self._version}.csv",
                self._src_dir / f"hemonc_synonyms_{self._version}.csv"
            )
            if not all((f.exists() for f in data_filenames)):
                self._download_data()
            self._src_files = data_filenames
            for file in self._src_files:
                assert file.exists()

    def _load_meta(self) -> None:
        """Add HemOnc metadata."""
        meta = {
            "data_license": "CC BY 4.0",
            "data_license_url": "https://creativecommons.org/licenses/by/4.0/legalcode",
            "version": self._version,
            "data_url": "https://dataverse.harvard.edu/dataset.xhtml?persistentId=doi:10.7910/DVN/9CY9C6",  # noqa: E501
            "rdp_url": None,
            "data_license_attributes": {
                "non_commercial": False,
                "share_alike": False,
                "attribution": True,
            },
        }
        assert SourceMeta(**meta)
        meta["src_name"] = SourceName.HEMONC.value
        self.database.metadata.put_item(Item=meta)

    def _get_concepts(self) -> Tuple[Dict, Dict, Dict]:
        """Get therapy, brand name, and disease concepts from concepts file.
        :return: Tuple of dicts mapping ID to object for each type of concept
        """
        therapies: RecordParams = {}  # hemonc id -> record
        brand_names: Dict[str, str] = {}  # hemonc id -> brand name
        conditions: Dict[str, str] = {}  # hemonc id -> condition name

        concepts_file = open(self._src_files[0], "r")
        concepts_reader = csv.reader(concepts_file)
        next(concepts_reader)  # skip header
        for row in concepts_reader:
            if row[6]:
                continue  # skip if deprecated/invalid

            row_type = row[2]
            if row_type == "Component":
                concept_id = f"{NamespacePrefix.HEMONC.value}:{row[3]}"
                therapies[row[3]] = {
                    "concept_id": concept_id,
                    "label": row[0],
                    "trade_names": [],
                    "aliases": [],
                    "xrefs": [],
                }
            elif row_type == "Brand Name":
                brand_names[row[3]] = row[0]
            elif row_type == "Condition":
                conditions[row[3]] = row[0]
        concepts_file.close()

        return therapies, brand_names, conditions

    @staticmethod
    def _id_to_yr(hemonc_id: str) -> str:
        """Get year from HemOnc ID corresponding to year concept.
        :param str hemonc_id: HemOnc ID to get year for
        :return: str representing year. Raises TypeError if HemOnc ID not valid.
        """
        id_int = int(hemonc_id)
        if id_int == 780:
            return "9999"
        elif id_int == 48349:
            return "2020"
        elif id_int == 5963:
            return "2021"
        elif id_int < 699 or id_int > 780:
            raise TypeError("ID not a valid HemOnc year concept")
        else:
            return str(id_int + 1240)

    def _get_rels(self, therapies: Dict, brand_names: Dict,
                  conditions: Dict) -> Dict:
        """Gather relations to provide associations between therapies, brand names,
        and conditions.

        :param dict therapies: mapping from IDs to therapy concepts
        :param dict brand_names: mapping from IDs to therapy brand names
        :param dict conditions: mapping from IDs to disease conditions
        :return: therapies dict updated with brand names and conditions
        """
        rels_file = open(self._src_files[1], "r")
        rels_reader = csv.reader(rels_file)
        next(rels_reader)  # skip header

        for row in rels_reader:
            rel_type = row[4]
            hemonc_id = row[0]
            record = therapies.get(hemonc_id)

            if record is None:
                continue  # skip non-drug items

            if rel_type == "Maps to":
                src_raw = row[3]
                if src_raw == "RxNorm":
                    xref = f"{NamespacePrefix.RXNORM.value}:{row[1]}"
                    record["xrefs"].append(xref)
                elif src_raw == "RxNorm Extension":
                    continue  # skip
                else:
                    logger.warning(f"Unrecognized `Maps To` source: {src_raw}")

            elif rel_type == "Has brand name":
                record["trade_names"].append(brand_names[row[1]])

            elif rel_type == "Was FDA approved yr":
                try:
                    year = self._id_to_yr(row[1])
                except TypeError:
                    logger.error(f"Failed parse of FDA approval year ID "
                                 f"{row[1]} for HemOnc ID {row[0]}")
                    continue
                if year == "9999":
                    logger.warning(f"HemOnc ID {row[0]} has FDA approval year"
                                   f" 9999")
                record["approval_ratings"] = [ApprovalRating.HEMONC_APPROVED.value]
                if "approval_year" in record:
                    record["approval_year"].append(year)
                else:
                    record["approval_year"] = [year]

            elif rel_type == "Has FDA indication":
                label = conditions[row[1]]
                norm_id = self._normalize_disease(label)
                hemonc_concept_id = f"{NamespacePrefix.HEMONC.value}:{row[1]}"
                indication = {
                    "disease_id": hemonc_concept_id,
                    "disease_label": label,
                    "normalized_disease_id": norm_id,
                    "supplemental_info": {"regulatory_body": "FDA"}
                }
                if "has_indication" in record:
                    record["has_indication"].append(indication)
                else:
                    record["has_indication"] = [indication]

        rels_file.close()
        return therapies

    def _get_synonyms(self, therapies: Dict) -> Dict:
        """Gather synonym entries and associate with therapy concepts.

        :param dict therapies: mapping of IDs to therapy objects
        :return: therapies dict with synonyms added as aliases
        """
        synonyms_file = open(self._src_files[2], "r")
        synonyms_reader = csv.reader(synonyms_file)
        next(synonyms_reader)
        for row in synonyms_reader:
            therapy_code = row[1]
            if therapy_code in therapies:
                therapy = therapies[therapy_code]
                alias = row[0]
                if alias != therapy.get("label"):
                    therapies[therapy_code]["aliases"].append(row[0])
        synonyms_file.close()
        return therapies

    def _transform_data(self) -> None:
        """Prepare dataset for loading into normalizer database."""
        therapies, brand_names, conditions = self._get_concepts()
        therapies = self._get_rels(therapies, brand_names, conditions)
        therapies = self._get_synonyms(therapies)

        for therapy in therapies.values():
            self._load_therapy(therapy)



"""ETL methods for the Drugs@FDA source."""
from typing import List, Optional
import json

import requests

from therapy import DownloadException, logger
from therapy.schemas import SourceMeta, SourceName, NamespacePrefix, ApprovalRating, \
    RecordParams
from therapy.etl.base import Base


class DrugsAtFDA(Base):
    """Class for Drugs@FDA ETL methods."""

    def _download_data(self) -> None:
        """Download source data from instance-provided source URL."""
        logger.info("Retrieving source data for Drugs@FDA")
        url = "https://download.open.fda.gov/drug/drugsfda/drug-drugsfda-0001-of-0001.json.zip"  # noqa: E501
        outfile_path = self._src_dir / f"drugsatfda_{self._version}.json"
        self._http_download(url, outfile_path, handler=self._zip_handler)
        logger.info("Successfully retrieved source data for Drugs@FDA")

    def get_latest_version(self) -> str:
        """Retrieve latest version of source data.
        :return: version as a str -- expected formatting is YYYY-MM-DD
        """
        r = requests.get("https://api.fda.gov/download.json")
        if r.status_code == 200:
            r_json = r.json()
            try:
                date = r_json["results"]["drug"]["drugsfda"]["export_date"]
            except KeyError:
                msg = "Unable to parse OpenFDA version API - check for breaking changes"
                logger.error(msg)
                raise DownloadException(msg)
            return date
        else:
            raise requests.HTTPError("Unable to retrieve version from FDA API")

    def _load_meta(self) -> None:
        """Add Drugs@FDA metadata."""
        meta = {
            "data_license": "CC0",
            "data_license_url": "https://creativecommons.org/publicdomain/zero/1.0/legalcode",  # noqa: E501
            "version": self._version,
            "data_url": "https://open.fda.gov/apis/drug/drugsfda/download/",
            "rdp_url": None,
            "data_license_attributes": {
                "non_commercial": False,
                "share_alike": False,
                "attribution": False,
            }
        }
        assert SourceMeta(**meta)
        meta["src_name"] = SourceName.DRUGSATFDA
        self.database.metadata.put_item(Item=meta)

    def _get_marketing_status_rating(self, products: List, concept_id: str)\
            -> Optional[str]:
        """Get approval status rating from products list.
        :param List products: list of individual FDA product objects
        :param str concept_id: FDA application concept ID, used in reporting error
            messages
        :return: approval_rating value if successful, None if ambiguous or unavailable
        """
        statuses_map = {
            "Discontinued": ApprovalRating.FDA_DISCONTINUED.value,
            "Prescription": ApprovalRating.FDA_PRESCRIPTION.value,
            "Over-the-counter": ApprovalRating.FDA_OTC.value,
            "None (Tentative Approval)": ApprovalRating.FDA_TENTATIVE.value,
        }
        statuses = [p["marketing_status"] for p in products]
        if not all([s == statuses[0] for s in statuses]):
            msg = (f"Application {concept_id} has inconsistent marketing "
                   f"statuses: {statuses}")
            logger.info(msg)
            return None
        else:
            return statuses_map.get(statuses[0])

    def _transform_data(self) -> None:
        """Prepare source data for loading into DB."""
        with open(self._src_file, "r") as f:
            data = json.load(f)["results"]

        for result in data:
            if "products" not in result:
                continue
            products = result["products"]

            app_no = result["application_number"]
            if app_no.startswith("NDA"):
                namespace = NamespacePrefix.DRUGSATFDA_NDA.value
            elif app_no.startswith("ANDA"):
                namespace = NamespacePrefix.DRUGSATFDA_ANDA.value
            else:
                # ignore biologics license applications (tentative)
                continue
            concept_id = f"{namespace}:{app_no.split('NDA')[-1]}"
            therapy: RecordParams = {"concept_id": concept_id}

            rating = self._get_marketing_status_rating(products, concept_id)
            if rating:
                therapy["approval_ratings"] = [rating]

            brand_names = [p["brand_name"] for p in products]

            aliases = []
            if "openfda" in result:
                openfda = result["openfda"]
                brand_name = openfda.get("brand_name")
                if brand_name:
                    brand_names += brand_name

                substances = openfda.get("substance_name", [])
                n_substances = len(set(substances))
                if n_substances > 1:
                    # if ambiguous, store all as aliases
                    msg = (f"Application {concept_id} has {n_substances} "
                           f"substance names: {substances}")
                    logger.debug(msg)
                    aliases += substances
                elif n_substances == 1:
                    therapy["label"] = substances[0]

                generics = openfda.get("generic_name", [])
                n_generic = len(set(generics))
                if n_generic > 1:
                    aliases += generics
                elif n_generic == 1:
                    if n_substances == 0:
                        # there are about 300 cases of this
                        therapy["label"] = generics[0]
                    else:
                        aliases.append(generics[0])

                therapy["associated_with"] = []
                unii = openfda.get("unii")
                if unii:
                    unii_items = [f"{NamespacePrefix.UNII.value}:{u}" for u in unii]
                    therapy["associated_with"] += unii_items  # type: ignore
                spl = openfda.get("spl_id")
                if spl:
                    spl_items = [f"{NamespacePrefix.SPL.value}:{s}" for s in spl]
                    therapy["associated_with"] += spl_items  # type: ignore
                ndc = openfda.get("product_ndc")
                if ndc:
                    ndc_items = [f"{NamespacePrefix.NDC.value}:{n}" for n in ndc]
                    therapy["associated_with"] += ndc_items  # type: ignore

                rxcui = openfda.get("rxcui")
                if rxcui:
                    therapy["xrefs"] = [f"{NamespacePrefix.RXNORM.value}:{r}"
                                        for r in rxcui]

            therapy["trade_names"] = brand_names
            therapy["aliases"] = aliases
            self._load_therapy(therapy)



"""This module provides methods for handling queries."""
import re
from typing import Callable, List, Dict, Set, Tuple, TypeVar, Union, Any, Optional
from urllib.parse import quote
from datetime import datetime
import json

from uvicorn.config import logger
from botocore.exceptions import ClientError

from therapy import SOURCES, PREFIX_LOOKUP, ITEM_TYPES, NAMESPACE_LUIS
from therapy.database import Database
from therapy.schemas import BaseNormalizationService, Drug, SourceMeta, MatchType, \
    ServiceMeta, HasIndication, SourcePriority, SearchService, NormalizationService, \
    NamespacePrefix, SourceName, TherapyDescriptor, UnmergedNormalizationService, \
    MatchesNormalized


NormService = TypeVar("NormService", bound=BaseNormalizationService)


class InvalidParameterException(Exception):
    """Exception for invalid parameter args provided by the user."""

    def __init__(self, message: str) -> None:
        """Create new instance

        :param str message: string describing the nature of the error
        """
        super().__init__(message)


class QueryHandler:
    """Class for normalizer management. Stores reference to database instance and
    normalizes query input.
    """

    def __init__(self, db_url: str = "", db_region: str = "us-east-2"):
        """Initialize Normalizer instance.

        :param str db_url: URL to database source.
        :param str db_region: AWS default region.
        """
        self.db = Database(db_url=db_url, region_name=db_region)

    def _emit_char_warnings(self, query_str: str) -> List[Dict]:
        """Emit warnings if query contains non breaking space characters.

        :param str query_str: query string
        :return: List of warnings (dicts)
        """
        warnings: List[Dict[str, str]] = []
        nbsp = re.search("\xa0|&nbsp;", query_str)
        if nbsp:
            warnings = [{
                "non_breaking_space_characters":
                    "Query contains non-breaking space characters"
            }]
            logger.warning(
                f"Query ({query_str}) contains non-breaking space characters."
            )
        return warnings

    def _fetch_meta(self, src_name: str) -> SourceMeta:
        """Fetch metadata for src_name.

        :param str src_name: name of source to get metadata for
        :return: SourceMeta object containing source metadata
        """
        if src_name in self.db.cached_sources.keys():
            return self.db.cached_sources[src_name]
        else:
            try:
                db_response = self.db.metadata.get_item(Key={"src_name": src_name})
            except ClientError as e:
                msg = e.response["Error"]["Message"]
                logger.error(msg)
                raise Exception(msg)
            try:
                response = SourceMeta(**db_response["Item"])
            except KeyError:
                msg = f"Metadata lookup failed for source {src_name}"
                logger.error(msg)
                raise Exception(msg)
            self.db.cached_sources[src_name] = response
            return response

    @staticmethod
    def _get_indication(indication_string: str) -> HasIndication:
        """Load indication data.
        :param str indication_string: dumped JSON string from db
        :return: complete HasIndication object
        """
        indication_values = json.loads(indication_string)
        return HasIndication(
            disease_id=indication_values[0],
            disease_label=indication_values[1],
            normalized_disease_id=indication_values[2],
            supplemental_info=indication_values[3]
        )

    def _add_record(self,
                    response: Dict[str, Dict],
                    item: Dict,
                    match_type: str) -> Tuple[Dict, str]:
        """Add individual record (i.e. Item in DynamoDB) to response object

        :param Dict[str, Dict] response: in-progress response object
        :param Dict item: Item retrieved from DynamoDB
        :param MatchType match_type: type of query match
        :return: Tuple containing updated response object, and string containing name of
            the source of the match
        """
        inds = item.get("has_indication")
        if inds:
            item["has_indication"] = [self._get_indication(i) for i in inds]

        drug = Drug(**item)
        src_name = item["src_name"]

        matches = response["source_matches"]
        if src_name not in matches.keys():
            pass
        elif matches[src_name] is None:
            matches[src_name] = {
                "match_type": MatchType[match_type.upper()],
                "records": [drug],
                "source_meta_": self._fetch_meta(src_name)
            }
        elif matches[src_name]["match_type"] == MatchType[match_type.upper()]:
            if drug.concept_id not in [r.concept_id for r
                                       in matches[src_name]["records"]]:
                matches[src_name]["records"].append(drug)

        return response, src_name

    def _fetch_records(self,
                       response: Dict[str, Dict],
                       concept_ids: Set[str],
                       match_type: str) -> Tuple[Dict, Set]:
        """Return matched Drug records as a structured response for a given collection
        of concept IDs.

        :param Dict[str, Dict] response: in-progress response object
        :param List[str] concept_ids: List of concept IDs to build from.  Should be all
            lower-case.
        :param str match_type: record should be assigned this type of match.
        :return: response Dict with records filled in via provided concept IDs, and Set
            of source names of matched records
        """
        matched_sources = set()
        for concept_id in concept_ids:
            try:
                match = self.db.get_record_by_id(concept_id.lower(),
                                                 case_sensitive=False)
                assert match, f"Unable to retrieve record for {concept_id}"
                (response, src) = self._add_record(response, match, match_type)
                matched_sources.add(src)
            except ClientError as e:
                logger.error(e.response["Error"]["Message"])

        return response, matched_sources

    def _fill_no_matches(self, resp: Dict[str, Any]) -> Dict:
        """Fill all empty source_matches slots with NO_MATCH results.

        :param Dict[str, Dict] resp: incoming response object
        :return: response object with empty source slots filled with NO_MATCH results
            and corresponding source metadata
        """
        for src_name in resp["source_matches"].keys():
            if resp["source_matches"][src_name] is None:
                resp["source_matches"][src_name] = {
                    "match_type": MatchType.NO_MATCH,
                    "records": [],
                    "source_meta_": self._fetch_meta(src_name)
                }
        return resp

    def _infer_namespace(self, query: str) -> Optional[Tuple[Dict, Dict]]:
        """Retrieve concept ID by inferring namespace. Attempts to match given query
        against known LUI patterns and performs concept ID lookup for all matches.
        :param str query: user-provided query string
        :return: Either tuple containing complete record and warnings if successful,
        or None if unsuccessful
        """
        inferred_records = []
        for pattern, source in NAMESPACE_LUIS:
            match = re.match(pattern, query)
            if match:
                if source == SourceName.DRUGSATFDA.value:
                    subspace, lui = match.groups()
                    namespace = f"drugsatfda.{subspace.lower()}"
                    inferred_id = f"{namespace}:{lui}"
                else:
                    namespace = NamespacePrefix[source.upper()].value
                    inferred_id = f"{namespace}:{query}"
                record = self.db.get_record_by_id(inferred_id, case_sensitive=False)
                if record:
                    inferred_records.append((record, namespace, inferred_id))
        if inferred_records:
            inferred_records.sort(key=lambda r: self._record_order(r[0]))
            return (
                inferred_records[0][0],
                {
                    "inferred_namespace": namespace,
                    "adjusted_query": inferred_records[0][2],
                    # probably not possible but just in case
                    "alternate_inferred_matches": [i[2] for i in inferred_records[1:]]
                }
            )
        else:
            return None

    def _check_concept_id(self, query: str, resp: Dict, sources: Set[str],
                          infer: bool = True) -> Tuple[Dict, Set]:
        """Check query for concept ID match. Should only find 0 or 1 matches.

        :param str query: search string
        :param Dict resp: in-progress response object to return to client
        :param Set[str] sources: remaining unmatched sources
        :param bool infer: if true, try to infer namespaces for IDs
        :return: Tuple with updated resp object and updated set of unmatched sources
        """
        records = []
        if infer:
            infer_response = self._infer_namespace(query)
            if infer_response:
                records.append(infer_response[0])
                resp["warnings"].append(infer_response[1])
        query_lower = query.lower()
        if [p for p in PREFIX_LOOKUP.keys() if query_lower.startswith(p)]:
            record = self.db.get_record_by_id(query, False)
            if record:
                records.append(record)
        for item in records:
            (resp, src_name) = self._add_record(resp, item,
                                                MatchType.CONCEPT_ID.name)
            sources = sources - {src_name}
        return resp, sources

    def _check_match_type(self, query: str, resp: Dict, sources: Set[str],
                          match_type: str) -> Tuple[Dict, Set]:
        """Check query for selected match type.
        :param str query: search string
        :param Dict resp: in-progress response object to return to client
        :param Set[str] sources: remaining unmatched sources
        :param str match_type: Match type to check for. Should be one of
        {'trade_name', 'label', 'alias', 'xref', 'associated_with'}
        :return: Tuple with updated resp object and updated set of unmatched
            sources
        """
        matches = self.db.get_records_by_type(query, match_type)
        if matches:
            concept_ids = {i["concept_id"] for i in matches}
            (resp, matched_srcs) = self._fetch_records(resp, concept_ids, match_type)
            sources = sources - matched_srcs
        return resp, sources

    def _response_keyed(self, query: str, sources: Set[str],
                        infer: bool = True) -> Dict:
        """Return response as dict where key is source name and value
        is a list of records. Corresponds to `keyed=true` API parameter.

        :param str query: string to match against
        :param Set[str] sources: sources to match from
        :param bool infer: if true, attempt to infer namespaces from IDs
        :return: completed response object to return to client
        """
        response: Dict[str, Union[None, str, List[Dict], Dict]] = {
            "query": query,
            "warnings": self._emit_char_warnings(query),
            "source_matches": {
                source: None for source in sources
            }
        }
        if query == "":
            response = self._fill_no_matches(response)
            return response
        query = query.strip()

        # check if concept ID match
        response, sources = self._check_concept_id(query, response, sources, infer)
        if len(sources) == 0:
            return response

        query = query.lower()
        for match_type in ITEM_TYPES.values():
            response, sources = self._check_match_type(query, response, sources,
                                                       match_type)
            if len(sources) == 0:
                return response

        # remaining sources get no match
        return self._fill_no_matches(response)

    def _response_list(self, query: str, sources: Set[str],
                       infer: bool = True) -> Dict:
        """Return response as list, where the first key-value in each item is the
        source name. Corresponds to `keyed=false` API parameter.

        :param str query: string to match against
        :param Set[str] sources: sources to match from
        :param bool infer: if true, attempt to infer namespaces from IDs
        :return: Completed response object to return to client
        """
        response_dict = self._response_keyed(query, sources, infer)
        source_list = []
        for src_name in response_dict["source_matches"].keys():
            src = {"source": src_name}
            to_merge = response_dict["source_matches"][src_name]
            src.update(to_merge)

            source_list.append(src)
        response_dict["source_matches"] = source_list

        return response_dict

    def search(self, query_str: str, keyed: bool = False, incl: str = "",
               excl: str = "", infer: bool = True) -> SearchService:
        """Fetch normalized therapy objects.

        :param str query_str: query, a string, to search for
        :param bool keyed: if true, return response as dict keying source names to
            source objects; otherwise, return list of source objects
        :param str incl: str containing comma-separated names of sources to use. Will
            exclude all other sources. Case-insensitive.
        :param str excl: str containing comma-separated names of source to exclude. Will
            include all other source. Case-insensitive.
        :param bool infer: if true, try to infer namespaces using known Local Unique
            Identifier patterns
        :return: dict containing all matches found in sources.
        :raises InvalidParameterException: if both incl and excl args are provided, or
            if invalid source names are given.
        """
        sources = dict()
        for k, v in SOURCES.items():
            if self.db.metadata.get_item(Key={"src_name": v}).get("Item"):
                sources[k] = v
        if not incl and not excl:
            query_sources = set(sources.values())
        elif incl and excl:
            detail = "Cannot request both source inclusions and exclusions."
            raise InvalidParameterException(detail)
        elif incl:
            req_sources = [n.strip() for n in incl.split(",")]
            invalid_sources = []
            query_sources = set()
            for source in req_sources:
                if source.lower() in sources.keys():
                    query_sources.add(sources[source.lower()])
                else:
                    invalid_sources.append(source)
            if invalid_sources:
                detail = f"Invalid source name(s): {invalid_sources}"
                raise InvalidParameterException(detail)
        else:
            req_exclusions = [n.strip() for n in excl.lower().split(",")]
            req_excl_dict = {r.lower(): r for r in req_exclusions}
            invalid_sources = []
            query_sources = set()
            for req_l, req in req_excl_dict.items():
                if req_l not in sources.keys():
                    invalid_sources.append(req)
            for src_l, src in sources.items():
                if src_l not in req_excl_dict.keys():
                    query_sources.add(src)
            if invalid_sources:
                detail = f"Invalid source name(s): {invalid_sources}"
                raise InvalidParameterException(detail)

        if keyed:
            response = self._response_keyed(query_str, query_sources, infer)
        else:
            response = self._response_list(query_str, query_sources, infer)

        response["service_meta_"] = ServiceMeta(
            response_datetime=datetime.now(),
        ).dict()
        return SearchService(**response)

    def _add_merged_meta(self, response: NormalizationService) -> NormalizationService:
        """Add source metadata to response object.

        :param NormalizationService response: in-progress response object
        :return: completed response object.
        """
        sources_meta = {}
        vod = response.therapy_descriptor

        xrefs = vod.xrefs or []  # type: ignore
        ids = [vod.therapy_id] + xrefs  # type: ignore
        for concept_id in ids:
            prefix = concept_id.split(":")[0]
            src_name = PREFIX_LOOKUP[prefix.lower()]
            if src_name not in sources_meta:
                sources_meta[src_name] = self._fetch_meta(src_name)
        response.source_meta_ = sources_meta  # type: ignore
        return response

    def _record_order(self, record: Dict) -> Tuple[int, str]:
        """Construct priority order for matching. Only called by sort().

        :param Dict record: individual record item in iterable to sort
        :return: tuple with rank value and concept ID
        """
        src = record["src_name"].upper()
        source_rank = SourcePriority[src]
        return source_rank, record["concept_id"]

    def _add_vod(self, response: NormalizationService, record: Dict, query: str,
                 match_type: MatchType) -> NormalizationService:
        """Format received DB record as VOD and update response object.
        :param NormalizationService response: in-progress response object
        :param Dict record: record as stored in DB
        :param str query: query string from user request
        :param MatchType match_type: type of match achieved
        :return: completed response object ready to return to user
        """
        vod = {
            "id": f"normalize.therapy:{quote(query.strip())}",
            "type": "TherapyDescriptor",
            "therapy_id": record["concept_id"],
            "label": record.get("label"),
            "extensions": [],
        }

        if "xrefs" in record:
            vod["xrefs"] = record["xrefs"]
        if "aliases" in record:
            vod["alternate_labels"] = record["aliases"]

        if any(filter(lambda f: f in record, ("approval_ratings",
                                              "approval_year",
                                              "has_indication"))):
            approv = {
                "type": "Extension",
                "name": "regulatory_approval",
                "value": {}
            }
            if "approval_ratings" in record:
                value = record.get("approval_ratings")
                if value:
                    approv["value"]["approval_ratings"] = value  # type: ignore
            if "approval_year" in record:
                value = record.get("approval_year")
                if value:
                    approv["value"]["approval_year"] = value  # type: ignore

            inds = record.get("has_indication", [])
            inds_list = []
            for ind_db in inds:
                indication = self._get_indication(ind_db)
                ind_value_obj: Dict[str, Optional[Union[str, List]]] = {
                    "id": indication.disease_id,
                    "type": "DiseaseDescriptor",
                    "label": indication.disease_label,
                    "disease_id": indication.normalized_disease_id,
                }
                if indication.supplemental_info:
                    ind_value_obj["extensions"] = [
                        {
                            "type": "Extension",
                            "name": k,
                            "value": v
                        }
                        for k, v in indication.supplemental_info.items()
                    ]
                inds_list.append(ind_value_obj)
            if inds_list:
                approv["value"]["has_indication"] = inds_list  # type: ignore
            vod["extensions"].append(approv)

        for field, name in (("trade_names", "trade_names"),
                            ("associated_with", "associated_with")):
            values = record.get(field)

            if values:
                vod["extensions"].append({
                    "type": "Extension",
                    "name": name,
                    "value": values
                })

        if not vod["extensions"]:
            del vod["extensions"]

        response.match_type = match_type
        response.therapy_descriptor = TherapyDescriptor(**vod)  # type: ignore
        response = self._add_merged_meta(response)
        return response

    def _resolve_merge(
        self, response: NormService, query: str, record: Dict,
        match_type: MatchType, callback: Callable
    ) -> NormService:
        """Given a record, return the corresponding normalized record

        :param NormalizationService response: in-progress response object
        :param str query: exact query as provided by user
        :param Dict record: record to retrieve normalized concept for
        :param MatchType match_type: type of match that returned these records
        :param Callable callback: response constructor method
        :return: Normalized response object
        """
        merge_ref = record.get("merge_ref")
        if merge_ref:
            # follow merge_ref
            merge = self.db.get_record_by_id(merge_ref, False, True)
            if merge is None:
                logger.error(f"Merge ref lookup failed for ref {record['merge_ref']} "
                             f"in record {record['concept_id']} from query `{query}`")
                return response
            else:
                return callback(response, merge, match_type)
        else:
            # record is sole member of concept group
            return callback(response, record, match_type)

    def _prepare_normalized_response(self, query: str) -> Dict[str, Any]:
        """Provide base response object for normalize endpoints.

        :param str query: user-provided query
        :return: basic normalization response boilerplate
        """
        return {
            "query": query,
            "match_type": MatchType.NO_MATCH,
            "warnings": self._emit_char_warnings(query),
            "service_meta_": ServiceMeta(response_datetime=datetime.now())
        }

    def _get_matches_by_type(self, query: str, match_type: str) -> List[Dict]:
        """Get matches list for match tier.
        :param str query: user-provided query
        :param str match_type: keyword of match type to check
        :return: List of records matching the query and match level
        """
        matching_refs = self.db.get_records_by_type(query, match_type)
        matching_records = [self.db.get_record_by_id(m["concept_id"], False)
                            for m in matching_refs]
        return sorted(matching_records, key=self._record_order)  # type: ignore

    def normalize(self, query: str, infer: bool = True) -> NormalizationService:
        """Return merged, normalized concept for given search term.

        :param str query: string to search against
        :param bool infer: if true, try to infer namespace for IDs
        :return: Normalized response object
        """
        # prepare basic response
        response = NormalizationService(**self._prepare_normalized_response(query))

        add_vod_curry = lambda res, rec, mat: self._add_vod(res, rec, query, mat)  # noqa: E501 E731
        return self._perform_normalized_lookup(response, query, infer, add_vod_curry)

    def _construct_drug_match(self, record: Dict) -> Drug:
        """Create individual Drug match for unmerged normalization endpoint.

        :param Dict record: record to add
        :return: completed Drug object
        """
        inds = record.get("has_indication")
        if inds:
            record["has_indication"] = [self._get_indication(i) for i in inds]
        return Drug(**record)

    def _add_normalized_records(self, response: UnmergedNormalizationService,
                                normalized_record: Dict,
                                match_type: MatchType) -> UnmergedNormalizationService:
        """Add individual records to unmerged normalize response.

        :param UnmergedNormalizationService response: in-progress response object
        :param Dict normalized_record: record associated with normalized concept,
            either merged or single identity
        :param MatchType match_type: type of match achieved
        :return: Completed response object
        """
        response.match_type = match_type
        response.normalized_concept_id = normalized_record["concept_id"]
        if normalized_record["item_type"] == "identity":
            record_source = SourceName[normalized_record["src_name"].upper()]
            response.source_matches[record_source] = MatchesNormalized(
                records=[self._construct_drug_match(normalized_record)],
                source_meta_=self._fetch_meta(record_source.value)
            )
        else:
            concept_ids = [normalized_record["concept_id"]] + \
                normalized_record.get("xrefs", [])
            for concept_id in concept_ids:
                record = self.db.get_record_by_id(concept_id, case_sensitive=False)
                if not record:
                    continue  # cover a few chemidplus edge cases
                record_source = SourceName[record["src_name"].upper()]
                drug = self._construct_drug_match(record)
                if record_source in response.source_matches:
                    response.source_matches[record_source].records.append(drug)
                else:
                    response.source_matches[record_source] = MatchesNormalized(
                        records=[drug],
                        source_meta_=self._fetch_meta(record_source.value)
                    )
        return response

    def _perform_normalized_lookup(self, response: NormService, query: str, infer: bool,
                                   response_builder: Callable) -> NormService:
        """Retrieve normalized concept, for use in normalization endpoints
        :param NormService response: in-progress response object
        :param str query: user-provided query
        :param bool infer: whether to try namespace inference
        :param Callable response_builder: response constructor callback method
        :return: completed service response object
        """
        if query == "":
            return response
        query_str = query.lower().strip()

        # check merged concept ID match
        record = self.db.get_record_by_id(query_str, case_sensitive=False, merge=True)
        if record:
            return response_builder(response, record, MatchType.CONCEPT_ID)

        # check concept ID match
        record = self.db.get_record_by_id(query_str, case_sensitive=False)
        if record:
            return self._resolve_merge(response, query, record, MatchType.CONCEPT_ID,
                                       response_builder)

        # check concept ID match with inferred namespace
        if infer:
            inferred_response = self._infer_namespace(query)
            if inferred_response:
                if response.warnings:
                    response.warnings.append(inferred_response[1])
                else:
                    response.warnings = [inferred_response[1]]
                return self._resolve_merge(response, query, inferred_response[0],
                                           MatchType.CONCEPT_ID,
                                           response_builder)

        # check other match types
        for match_type in ITEM_TYPES.values():
            matching_records = self._get_matches_by_type(query_str, match_type)

            # attempt merge ref resolution until successful
            for match in matching_records:
                assert match is not None
                record = self.db.get_record_by_id(match["concept_id"], False)
                if record:
                    match_type_value = MatchType[match_type.upper()]
                    return self._resolve_merge(response, query, record,
                                               match_type_value,
                                               response_builder)

        return response

    def normalize_unmerged(
        self, query: str, infer: bool = True
    ) -> UnmergedNormalizationService:
        """Return all source records under the normalized concept for the provided
        query string.

        :param str query: string to search against
        :param bool infer: if true, try to infer namespace for IDs
        :return: Normalized response object
        """
        response = UnmergedNormalizationService(
            source_matches={},
            **self._prepare_normalized_response(query)
        )
        return self._perform_normalized_lookup(response, query, infer,
                                               self._add_normalized_records)



"""This module defines the Wikidata ETL methods."""
import json
import logging
import datetime
from typing import Dict, Any

from wikibaseintegrator.wbi_helpers import execute_sparql_query

from therapy import XREF_SOURCES, DownloadException
from therapy.schemas import SourceName, NamespacePrefix, RecordParams, SourceMeta
from therapy.etl.base import Base

logger = logging.getLogger("therapy")
logger.setLevel(logging.DEBUG)

# Translate Wikidata keys to standardized namespaces
NAMESPACES = {
    "casRegistry": NamespacePrefix.CASREGISTRY.value,
    "ChemIDplus": NamespacePrefix.CHEMIDPLUS.value,
    "pubchemCompound": NamespacePrefix.PUBCHEMCOMPOUND.value,
    "pubchemSubstance": NamespacePrefix.PUBCHEMSUBSTANCE.value,
    "chembl": NamespacePrefix.CHEMBL.value,
    "rxnorm": NamespacePrefix.RXNORM.value,
    "drugbank": NamespacePrefix.DRUGBANK.value,
    "wikidata": NamespacePrefix.WIKIDATA.value,
}

# Provide standard concept ID prefixes
ID_PREFIXES = {
    "wikidata": "Q",
    "chembl": "CHEMBL",
    "drugbank": "DB",
    "ncit": "C",
}

SPARQL_QUERY = """
    SELECT ?item ?itemLabel ?casRegistry ?pubchemCompound
           ?pubchemSubstance ?chembl
           ?rxnorm ?drugbank ?alias WHERE {
      ?item (wdt:P31/(wdt:P279*)) wd:Q12140.
      OPTIONAL {
        ?item skos:altLabel ?alias.
        FILTER((LANG(?alias)) = \"en\")
      }
      OPTIONAL { ?item p:P231 ?wds1.
                 ?wds1 ps:P231 ?casRegistry.
               }
      OPTIONAL { ?item p:P662 ?wds2.
                 ?wds2 ps:P662 ?pubchemCompound.
               }
      OPTIONAL { ?item p:P2153 ?wds3.
                 ?wds3 ps:P2153 ?pubchemSubstance.
               }
      OPTIONAL { ?item p:P592 ?wds4.
                 ?wds4 ps:P592 ?chembl
               }
      OPTIONAL { ?item p:P3345 ?wds5.
                 ?wds5 ps:P3345 ?rxnorm.
               }
      OPTIONAL { ?item p:P715 ?wds6.
                 ?wds6 ps:P715 ?drugbank
               }
      SERVICE wikibase:label {
        bd:serviceParam wikibase:language \"[AUTO_LANGUAGE],en\".
      }
    }
"""


class Wikidata(Base):
    """Class for Wikidata ETL methods."""

    def _download_data(self) -> None:
        """Download latest Wikidata source dump."""
        logger.info("Retrieving source data for Wikidata")
        query_results = execute_sparql_query(SPARQL_QUERY)
        if query_results is None:
            raise DownloadException("Wikidata SPARQL query returned no results")
        else:
            data = query_results["results"]["bindings"]

        transformed_data = []
        for item in data:
            params: RecordParams = {}
            for attr in item:
                params[attr] = item[attr]["value"]
            transformed_data.append(params)
        with open(f"{self._src_dir}/wikidata_{self._version}.json", "w+") as f:
            json.dump(transformed_data, f)
        logger.info("Successfully retrieved source data for Wikidata")

    def get_latest_version(self) -> str:
        """Wikidata is constantly, immediately updated, so source data has no strict
        versioning. We use the current date as a pragmatic way to indicate the version.
        """
        return datetime.datetime.today().strftime("%Y-%m-%d")

    def _load_meta(self) -> None:
        """Add Wikidata metadata."""
        metadata = SourceMeta(
            src_name=SourceName.WIKIDATA.value,
            data_license="CC0 1.0",
            data_license_url="https://creativecommons.org/publicdomain/zero/1.0/",
            version=self._version,
            data_url=None,
            rdp_url=None,
            data_license_attributes={
                "non_commercial": False,
                "share_alike": False,
                "attribution": False
            }
        )
        params = dict(metadata)
        params["src_name"] = SourceName.WIKIDATA.value
        self.database.metadata.put_item(Item=params)

    def _transform_data(self) -> None:
        """Transform the Wikidata source data."""
        with open(self._src_file, "r") as f:
            records = json.load(f)

            items: Dict[str, Any] = dict()

            for record in records:
                record_id = record["item"].split("/")[-1]
                concept_id = f"{NamespacePrefix.WIKIDATA.value}:{record_id}"
                if concept_id not in items.keys():
                    item: Dict[str, Any] = {"concept_id": concept_id}

                    xrefs = list()
                    associated_with = list()
                    for key in NAMESPACES.keys():
                        if key in record.keys():
                            ref = record[key]

                            if key.upper() == "CASREGISTRY":
                                key = SourceName.CHEMIDPLUS.value

                            if key.upper() in XREF_SOURCES:
                                if key != "chembl":
                                    prefix = ID_PREFIXES.get(key.lower(), "")
                                    fmted_xref = f"{NAMESPACES[key]}:{prefix}{ref}"
                                else:
                                    fmted_xref = f"{NAMESPACES[key]}:{ref}"
                                xrefs.append(fmted_xref)
                            else:
                                fmted_assoc = f"{NAMESPACES[key]}:" \
                                              f"{ref}"
                                associated_with.append(fmted_assoc)
                    item["xrefs"] = xrefs
                    item["associated_with"] = associated_with
                    if "itemLabel" in record.keys():
                        item["label"] = record["itemLabel"]
                    items[concept_id] = item
                if "alias" in record.keys():
                    if "aliases" in items[concept_id].keys():
                        items[concept_id]["aliases"].append(record["alias"])
                    else:
                        items[concept_id]["aliases"] = [record["alias"]]

        for item in items.values():
            self._load_therapy(item)



"""This module contains data models for representing VICC therapy records."""
from typing import List, Literal, Optional, Dict, Union, Any, Type, Set
from enum import Enum, IntEnum
from datetime import datetime

from ga4gh.vrsatile.pydantic import return_value
from ga4gh.vrsatile.pydantic.vrs_models import CURIE
from ga4gh.vrsatile.pydantic.vrsatile_models import ValueObjectDescriptor
from pydantic import BaseModel, StrictBool, validator

from therapy.version import __version__

# Working structure for object in preparation for upload to DB
RecordParams = Dict[str, Union[List, Set, str, Dict[str, Any]]]


class ApprovalRating(str, Enum):
    """Define string constraints for approval rating attribute. This reflects a drug's
    regulatory approval status as evaluated by an individual source. We opted to retain
    each individual source's rating as a distinct enum value after finding that some
    sources disagreed on the true value (either because of differences in scope like
    US vs EU/other regulatory arenas, old or conflicting data, or other reasons). Value
    descriptions are provided below from each listed source.

    ChEMBL:
     - CHEMBL_PHASE_0: "Research: The compound has not yet reached clinical trials
    (preclinical/research compound)"
     - CHEMBL_PHASE_1: "The compound has reached Phase I clinical trials (safety
    studies, usually with healthy volunteers)"
     - CHEMBL_PHASE_2: "The compound has reached Phase II clinical trials (preliminary
    studies of effectiveness)"
     - CHEMBL_PHASE_3: "The compound has reached Phase III clinical trials (larger
    studies of safety and effectiveness)"
     - CHEMBL_PHASE_4: The compound has been approved in at least one country or area."
     - CHEMBL_WITHDRAWN: "A withdrawn drug is an approved drug contained in a medicinal
    product that subsequently had been removed from the market. The reasons for
    withdrawal may include toxicity, lack of efficacy, or other reasons such as an
    unfavorable risk-to-benefit ratio following approval and marketing of the drug.
    ChEMBL considers an approved drug to be withdrawn only if all medicinal products
    that contain the drug as an active ingredient have been withdrawn from one (or more)
    regions of the world. Note that all medicinal products for a drug can be withdrawn
    in one region of the world while still being marketed in other jurisdictions."
    https://pubs.acs.org/doi/10.1021/acs.chemrestox.0c00296

    Drugs@FDA:
     - FDA_PRESCRIPTION: "A prescription drug product requires a doctor's authorization
    to purchase."
     - FDA_OTC: "FDA defines OTC drugs as safe and effective for use by the general
    public without a doctor's prescription."
     - FDA_DISCONTINUED: "approved products that have never been marketed, have been
    discontinued from marketing, are for military use, are for export only, or have had
    their approvals withdrawn for reasons other than safety or efficacy after being
    discontinued from marketing"
     - FDA_TENTATIVE: "If a generic drug product is ready for approval before the
    expiration of any patents or exclusivities accorded to the reference listed drug
    product, FDA issues a tentative approval letter to the applicant.  FDA delays final
    approval of the generic drug product until all patent or exclusivity issues have
    been resolved."
    https://www.fda.gov/drugs/drug-approvals-and-databases/drugsfda-glossary-terms

    HemOnc.org:
    - HEMONC_APPROVED: Inferred by us if "Was FDA Approved Yr" property is present
    (described as "Year of FDA approval")
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6697579/

    Guide to Pharmacology:
    - GTOPDB_APPROVED: "Indicates pharmacologicaly active substances, specified by their
    INNs, that have been approved for clinical use by a regulatory agency, typically the
    FDA, EMA or in Japan. This classification persists regardless of whether the drug
    may later have been withdrawn or discontinued. (N.B. in some cases the information
    on approval status was obtained indirectly via databases such as Drugbank.)"
    - GTOPDB_WITHDRAWN: "No longer approved for its original clinical use, as notified
    by the FDA, typically as a consequence of safety or side effect issues."
    https://www.guidetopharmacology.org/helpPage.jsp

    RxNorm:
    - RXNORM_PRESCRIBABLE: "The RxNorm Current Prescribable Content is a subset of
    currently prescribable drugs found in RxNorm. We intend it to be an approximation of
    the prescription drugs currently marketed in the US. The subset also includes many
    over-the-counter drugs."
    https://www.nlm.nih.gov/research/umls/rxnorm/docs/prescribe.html
    """

    CHEMBL_0 = "chembl_phase_0"
    CHEMBL_1 = "chembl_phase_1"
    CHEMBL_2 = "chembl_phase_2"
    CHEMBL_3 = "chembl_phase_3"
    CHEMBL_4 = "chembl_phase_4"
    CHEMBL_WITHDRAWN = "chembl_withdrawn"
    FDA_OTC = "fda_otc"
    FDA_PRESCRIPTION = "fda_prescription"
    FDA_DISCONTINUED = "fda_discontinued"
    FDA_TENTATIVE = "fda_tentative"
    HEMONC_APPROVED = "hemonc_approved"
    GTOPDB_APPROVED = "gtopdb_approved"
    GTOPDB_WITHDRAWN = "gtopdb_withdrawn"
    RXNORM_PRESCRIBABLE = "rxnorm_prescribable"


class HasIndication(BaseModel):
    """Data regarding disease indications from regulatory bodies."""

    disease_id: str
    disease_label: str
    normalized_disease_id: Optional[str]
    supplemental_info: Optional[Dict[str, str]] = None

    class Config:
        """Configure HasIndication class"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type["HasIndication"]) -> None:
            """Configure OpenAPI schema"""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = [
                {
                    "disease_id": "mesh:D016778",
                    "disease_label": "Malaria, Falciparum",
                    "normalized_disease_id": "ncit:C34798",
                    "supplemental_info": {
                        "chembl_max_phase_for_ind": "chembl_phase_2"
                    }
                },
                {
                    "disease_id": "hemonc:634",
                    "disease_label": "Myelodysplastic syndrome",
                    "normalized_disease_id": "ncit:C3247",
                    "supplemental_info": {
                        "regulatory_body": "FDA"
                    }
                }
            ]


class Drug(BaseModel):
    """A pharmacologic substance used to treat a medical condition."""

    concept_id: str
    label: Optional[str] = None
    aliases: Optional[List[str]] = []
    trade_names: Optional[List[str]] = []
    xrefs: Optional[List[str]] = []
    associated_with: Optional[List[str]] = []
    approval_ratings: Optional[List[ApprovalRating]] = None
    approval_year: Optional[List[str]] = []
    has_indication: Optional[List[HasIndication]] = []

    class Config:
        """Configure Drug class"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type["Drug"]) -> None:
            """Configure OpenAPI schema"""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "label": "CISPLATIN",
                "concept_id": "chembl:CHEMBL11359",
                "aliases": [
                    "Cisplatin",
                    "Cis-Platinum II",
                    "Cisplatinum",
                    "cis-diamminedichloroplatinum(II)",
                    "CIS-DDP",
                    "INT-230-6 COMPONENT CISPLATIN",
                    "INT230-6 COMPONENT CISPLATIN",
                    "NSC-119875",
                    "Platinol",
                    "Platinol-Aq"
                ],
                "xrefs": [],
                "associated_with": None,
                "approval_ratings": "approved",
                "approval_year": [],
                "has_indication": [],
                "trade_names": ["PLATINOL", "PLATINOL-AQ", "CISPLATIN"]
            }


class MatchType(IntEnum):
    """Define string constraints for use in Match Type attributes."""

    CONCEPT_ID = 100
    LABEL = 80
    TRADE_NAME = 80
    ALIAS = 60
    XREF = 60
    ASSOCIATED_WITH = 60
    FUZZY_MATCH = 20
    NO_MATCH = 0


class SourcePriority(IntEnum):
    """Define constraints for Source Priority Rankings."""

    RXNORM = 1
    NCIT = 2
    HEMONC = 3
    DRUGBANK = 4
    DRUGSATFDA = 5
    GUIDETOPHARMACOLOGY = 6
    CHEMBL = 7
    CHEMIDPLUS = 8
    WIKIDATA = 9


class SourceName(str, Enum):
    """Define string constraints to ensure consistent capitalization."""

    WIKIDATA = "Wikidata"
    CHEMBL = "ChEMBL"
    NCIT = "NCIt"
    DRUGBANK = "DrugBank"
    CHEMIDPLUS = "ChemIDplus"
    RXNORM = "RxNorm"
    HEMONC = "HemOnc"
    DRUGSATFDA = "DrugsAtFDA"
    DRUGSATFDA_NDA = DRUGSATFDA
    DRUGSATFDA_ANDA = DRUGSATFDA
    GUIDETOPHARMACOLOGY = "GuideToPHARMACOLOGY"


class NamespacePrefix(Enum):
    """Define string constraints for namespace prefixes on concept IDs."""

    ATC = "atc"  # Anatomical Therapeutic Chemical Classification System
    CHEBI = "CHEBI"
    CHEMBL = "chembl"
    CHEMIDPLUS = "chemidplus"
    CASREGISTRY = CHEMIDPLUS
    CVX = "cvx"  # Vaccines Administered
    DRUGBANK = "drugbank"
    DRUGCENTRAL = "drugcentral"
    DRUGSATFDA_ANDA = "drugsatfda.anda"
    DRUGSATFDA_NDA = "drugsatfda.nda"
    HEMONC = "hemonc"
    INCHIKEY = "inchikey"
    IUPHAR_LIGAND = "iuphar.ligand"
    GUIDETOPHARMACOLOGY = IUPHAR_LIGAND
    MMSL = "mmsl"  # Multum MediSource Lexicon
    MSH = "mesh"  # Medical Subject Headings
    NCIT = "ncit"
    NDC = "ndc"  # National Drug Code
    PUBCHEMCOMPOUND = "pubchem.compound"
    PUBCHEMSUBSTANCE = "pubchem.substance"
    RXNORM = "rxcui"
    SPL = "spl"  # Structured Product Labeling
    UMLS = "umls"
    UNII = "unii"
    UNIPROT = "uniprot"
    USP = "usp"  # USP Compendial Nomenclature
    VANDF = "vandf"  # Veterans Health Administration National Drug File
    WIKIDATA = "wikidata"


class DataLicenseAttributes(BaseModel):
    """Define constraints for data license attributes."""

    non_commercial: StrictBool
    share_alike: StrictBool
    attribution: StrictBool


class ItemTypes(str, Enum):
    """Item types used in DynamoDB."""

    # Must be in descending MatchType order.
    LABEL = "label"
    TRADE_NAMES = "trade_name"
    ALIASES = "alias"
    XREFS = "xref"
    ASSOCIATED_WITH = "associated_with"


class SourceMeta(BaseModel):
    """Metadata for a given source to return in response object."""

    data_license: str
    data_license_url: str
    version: str
    data_url: Optional[str]
    rdp_url: Optional[str]
    data_license_attributes: Dict[str, StrictBool]

    class Config:
        """Configure OpenAPI schema"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type["SourceMeta"]) -> None:
            """Configure OpenAPI schema"""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "data_license": "CC BY-SA 3.0",
                "data_license_url":
                    "https://creativecommons.org/licenses/by-sa/3.0/",
                "version": "27",
                "data_url":
                    "http://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_27/",  # noqa: E501
                "rdp_url": "http://reusabledata.org/chembl.html",
                "data_license_attributes": {
                    "non_commercial": False,
                    "share_alike": True,
                    "attribution": True
                }
            }


class MatchesKeyed(BaseModel):
    """Container for matching information from an individual source.
    Used when matches are requested as an object, not an array.
    """

    match_type: MatchType
    records: List[Drug]
    source_meta_: SourceMeta

    class Config:
        """Configure OpenAPI schema"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type["MatchesKeyed"]) -> None:
            """Configure OpenAPI schema"""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "match_type": 0,
                "records": [],
                "source_meta_": {
                    "data_license": "CC BY-SA 3.0",
                    "data_license_url":
                        "https://creativecommons.org/licenses/by-sa/3.0/",
                    "version": "27",
                    "data_url":
                        "http://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_27/",  # noqa: E501
                    "rdp_url": "http://reusabledata.org/chembl.html",
                    "data_license_attributes": {
                        "non_commercial": False,
                        "share_alike": True,
                        "attribution": True
                    }
                },
            }


class MatchesListed(BaseModel):
    """Container for matching information from an individual source.
    Used when matches are requested as an array, not an object.
    """

    source: SourceName
    match_type: MatchType
    records: List[Drug]
    source_meta_: SourceMeta

    class Config:
        """Configure openAPI schema"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type["MatchesListed"]) -> None:
            """Configure OpenAPI schema"""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "source": "ChEMBL",
                "match_type": 0,
                "records": [],
                "source_meta_": {
                    "data_license": "CC BY-SA 3.0",
                    "data_license_url":
                        "https://creativecommons.org/licenses/by-sa/3.0/",
                    "version": "27",
                    "data_url":
                        "http://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_27/",  # noqa: E501
                    "rdp_url": "http://reusabledata.org/chembl.html",
                    "data_license_attributes": {
                        "non_commercial": False,
                        "share_alike": True,
                        "attribution": True
                    }
                },
            }


class TherapyDescriptor(ValueObjectDescriptor):
    """Create therapy descriptor model"""

    type: Literal["TherapyDescriptor"] = "TherapyDescriptor"
    therapy_id: CURIE

    _get_therapy_id_val = validator("therapy_id", allow_reuse=True)(return_value)


class ServiceMeta(BaseModel):
    """Metadata regarding the therapy-normalization service."""

    name = "thera-py"
    version = __version__
    response_datetime: datetime
    url = "https://github.com/cancervariants/therapy-normalization"

    class Config:
        """Configure OpenAPI schema"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type["ServiceMeta"]) -> None:
            """Configure OpenAPI schema"""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "name": "thera-py",
                "version": "0.1.0",
                "response_datetime": "2021-04-05T16:44:15.367831",
                "url": "https://github.com/cancervariants/therapy-normalization"
            }


class MatchesNormalized(BaseModel):
    """Matches associated with normalized concept from a single source."""

    records: List[Drug]
    source_meta_: SourceMeta

    class Config:
        """Configure OpenAPI schema"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["MatchesNormalized"]) -> None:
            """Configure OpenAPI schema"""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)


class BaseNormalizationService(BaseModel):
    """Base method providing shared attributes to Normalization service classes."""

    query: str
    warnings: Optional[List[Dict]]
    match_type: MatchType
    service_meta_: ServiceMeta


class UnmergedNormalizationService(BaseNormalizationService):
    """Response providing source records corresponding to normalization of user query.
    Enables retrieval of normalized concept while retaining sourcing for accompanying
    attributes.
    """

    normalized_concept_id: Optional[CURIE]
    source_matches: Dict[SourceName, MatchesNormalized]

    class Config:
        """Configure OpenAPI schema"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["UnmergedNormalizationService"]) -> None:
            """Configure OpenAPI schema example"""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "query": "L745870",
                "warnings": [],
                "match_type": 80,
                "service_meta_": {
                    "response_datetime": "2022-04-22T11:40:18.921859",
                    "name": "thera-py",
                    "version": "0.3.4",
                    "url": "https://github.com/cancervariants/therapy-normalization"
                },
                "normalized_concept_id": "iuphar.ligand:3303",
                "source_matches": {
                    "GuideToPHARMACOLOGY": {
                        "records": [
                            {
                                "concept_id": "iuphar.ligand:3303",
                                "label": "L745870",
                                "aliases": [
                                    "L-745,870",
                                    "L 745870",
                                    "3-[[4-(4-chlorophenyl)piperazin-1-yl]methyl]-1H-pyrrolo[2,3-b]pyridine"  # noqa: E501
                                ],
                                "trade_names": [],
                                "xrefs": [
                                    "chemidplus:158985-00-3",
                                    "chembl:CHEMBL267014"
                                ],
                                "associated_with": [
                                    "pubchem.substance:178100340",
                                    "pubchem.compound:5311200",
                                    "inchikey:OGJGQVFWEPNYSB-UHFFFAOYSA-N"
                                ],
                                "approval_ratings": None,
                                "approval_year": [],
                                "has_indication": []
                            }
                        ],
                        "source_meta_": {
                            "data_license": "CC BY-SA 4.0",
                            "data_license_url": "https://creativecommons.org/licenses/by-sa/4.0/",  # noqa: E501
                            "version": "2021.4",
                            "data_url": "https://www.guidetopharmacology.org/download.jsp",  # noqa: E501
                            "rdp_url": None,
                            "data_license_attributes": {
                                "non_commercial": False,
                                "share_alike": True,
                                "attribution": True
                            }
                        }
                    },
                    "ChEMBL": {
                        "records": [
                            {
                                "concept_id": "chembl:CHEMBL267014",
                                "label": "L-745870",
                                "aliases": [],
                                "trade_names": [],
                                "xrefs": [],
                                "associated_with": [],
                                "approval_ratings": [
                                    "chembl_phase_0"
                                ],
                                "approval_year": [],
                                "has_indication": []
                            }
                        ],
                        "source_meta_": {
                            "data_license": "CC BY-SA 3.0",
                            "data_license_url": "https://creativecommons.org/licenses/by-sa/3.0/",  # noqa: E501
                            "version": "29",
                            "data_url": "ftp://ftp.ebi.ac.uk/pub/databases/chembl/ChEMBLdb/releases/chembl_29",  # noqa: E501
                            "rdp_url": "http://reusabledata.org/chembl.html",
                            "data_license_attributes": {
                                "non_commercial": False,
                                "share_alike": True,
                                "attribution": True
                            }
                        }
                    }
                }
            }


class NormalizationService(BaseNormalizationService):
    """Response containing one or more merged records and source data."""

    therapy_descriptor: Optional[TherapyDescriptor]
    source_meta_: Optional[Dict[SourceName, SourceMeta]]

    class Config:
        """Configure OpenAPI schema"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any],
                         model: Type["NormalizationService"]) -> None:
            """Configure OpenAPI schema"""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "query": "cisplatin",
                "warnings": None,
                "match_type": 80,
                "therapy_descriptor": {
                    "id": "normalize.therapy:cisplatin",
                    "type": "TherapyDescriptor",
                    "therapy_id": "rxcui:2555",
                    "label": "cisplatin",
                    "xrefs": [
                        "ncit:C376", "chemidplus:15663-27-1",
                        "wikidata:Q412415"
                    ],
                    "alternate_labels": [
                        "CIS-DDP", "cis Platinum", "DDP",
                        "Dichlorodiammineplatinum",
                        "1,2-Diaminocyclohexaneplatinum II citrate",
                        "CISplatin",
                        "cis Diamminedichloroplatinum",
                        "CDDP",
                        "Diamminodichloride, Platinum",
                        "cis-Dichlorodiammineplatinum(II)",
                        "cis-Platinum",
                        "cis-diamminedichloroplatinum(II)",
                        "cis-Diamminedichloroplatinum(II)",
                        "Cis-DDP",
                        "cis-Diamminedichloroplatinum",
                        "cis-Diaminedichloroplatinum",
                        "Platinol-AQ", "Platinol",
                        "Platinum Diamminodichloride"
                    ],
                    "extensions": [
                        {
                            "type": "Extension",
                            "name": "trade_names",
                            "value": [
                                "Platinol", "Cisplatin"
                            ]
                        },
                        {
                            "type": "Extension",
                            "name": "associated_with",
                            "value": [
                                "atc:L01XA01",
                                "mmsl:4456",
                                "chebi:CHEBI:27899",
                                "pubchem.compound:5702198",
                                "umls:C0008838",
                                "usp:m17910",
                                "fda:Q20Q21Q62J",
                                "mmsl:d00195",
                                "mthspl:Q20Q21Q62J",
                                "mmsl:31747",
                                "mesh:D002945",
                                "vandf:4018139"
                            ]
                        }
                    ]
                },
                "source_meta_": {
                    "RxNorm": {
                        "data_license": "UMLS Metathesaurus",
                        "data_license_url": "https://www.nlm.nih.gov/research/umls/rxnorm/docs/termsofservice.html",  # noqa: E501
                        "version": "20210104",
                        "data_url": "https://www.nlm.nih.gov/research/umls/rxnorm/docs/rxnormfiles.html",  # noqa: E501
                        "rdp_url": None,
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": True,
                            "share_alike": False
                        }
                    },
                    "NCIt": {
                        "data_license": "CC BY 4.0",
                        "data_license_url": "https://creativecommons.org/licenses/by/4.0/legalcode",  # noqa: E501
                        "version": "20.09d",
                        "data_url": "https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/archive/20.09d_Release/",  # noqa: E501
                        "rdp_url": "http://reusabledata.org/ncit.html",
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": True,
                            "share_alike": False
                        }
                    },
                    "ChemIDplus": {
                        "data_license": "custom",
                        "data_license_url": "https://www.nlm.nih.gov/databases/download/terms_and_conditions.html",  # noqa: E501
                        "version": "20200327", "data_url": "ftp://ftp.nlm.nih.gov/nlmdata/.chemidlease/",  # noqa: E501
                        "rdp_url": None,
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": True,
                            "share_alike": False
                        }
                    },
                    "Wikidata": {
                        "data_license": "CC0 1.0",
                        "data_license_url": "https://creativecommons.org/publicdomain/zero/1.0/",  # noqa: E501
                        "version": "20200812",
                        "data_url": None,
                        "rdp_url": None,
                        "data_license_attributes": {
                            "non_commercial": False,
                            "attribution": False,
                            "share_alike": False
                        }
                    }
                },
                "service_meta_": {
                    "name": "thera-py",
                    "version": "0.1.0",
                    "response_datetime": "2021-04-05T16:44:15.367831",
                    "url": "https://github.com/cancervariants/therapy-normalization"
                }
            }


class SearchService(BaseModel):
    """Core response schema containing matches for each source"""

    query: str
    warnings: Optional[List[Dict]]
    source_matches: Union[Dict[SourceName, MatchesKeyed], List[MatchesListed]]
    service_meta_: ServiceMeta

    class Config:
        """Enables orm_mode"""

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type["SearchService"]) -> None:
            """Configure OpenAPI schema"""
            if "title" in schema.keys():
                schema.pop("title", None)
            for prop in schema.get("properties", {}).values():
                prop.pop("title", None)
            schema["example"] = {
                "query": "cisplatin",
                "warnings": None,
                "source_matches": [
                    {
                        "source": "ChemIDplus",
                        "match_type": 80,
                        "records": [
                            {
                                "label": "Cisplatin",
                                "concept_id": "chemidplus:15663-27-1",
                                "aliases": [
                                    "cis-Diaminedichloroplatinum",
                                    "1,2-Diaminocyclohexaneplatinum II citrate"
                                ],
                                "xrefs": ["drugbank:DB00515"],
                                "associated_with": ["fda:Q20Q21Q62J"],
                                "approval_ratings": None,
                                "trade_names": []
                            }
                        ],
                        "source_meta_": {
                            "data_license": "custom",
                            "data_license_url": "https://www.nlm.nih.gov/databases/download/terms_and_conditions.html",  # noqa: E501
                            "version": "20210204",
                            "data_url": "ftp://ftp.nlm.nih.gov/nlmdata/.chemidlease/",
                            "rdp_url": None,
                            "data_license_attributes": {
                                "non_commercial": False,
                                "attribution": True,
                                "share_alike": False
                            }
                        }
                    },
                    {
                        "source": "RxNorm",
                        "match_type": 80,
                        "records": [
                            {
                                "label": "cisplatin",
                                "concept_id": "rxcui:2555",
                                "aliases": [
                                    "cis-Dichlorodiammineplatinum(II)",
                                    "Platinum Diamminodichloride",
                                    "cis Diamminedichloroplatinum",
                                    "cis-diamminedichloroplatinum(II)",
                                    "cis-Diamminedichloroplatinum",
                                    "cis Platinum",
                                    "CDDP",
                                    "Dichlorodiammineplatinum",
                                    "cis-Platinum",
                                    "CISplatin",
                                    "cis-Diamminedichloroplatinum(II)",
                                    "Cis-DDP",
                                    "DDP",
                                    "Diamminodichloride, Platinum"
                                ],
                                "xrefs": [
                                    "drugbank:DB00515",
                                    "drugbank:DB12117"
                                ],
                                "associated_with": [
                                    "usp:m17910",
                                    "vandf:4018139",
                                    "mesh:D002945",
                                    "mthspl:Q20Q21Q62J",
                                    "mmsl:d00195",
                                    "atc:L01XA01",
                                    "mmsl:31747",
                                    "mmsl:4456"
                                ],
                                "approval_ratings": ["rxnorm_prescribable"],
                                "trade_names": [
                                    "Cisplatin",
                                    "Platinol"
                                ]
                            }
                        ],
                        "source_meta_": {
                            "data_license": "UMLS Metathesaurus",
                            "data_license_url": "https://www.nlm.nih.gov/research/umls/rxnorm/docs/termsofservice.html",  # noqa: E501
                            "version": "20210104",
                            "data_url": "https://www.nlm.nih.gov/research/umls/rxnorm/docs/rxnormfiles.html",  # noqa: E501
                            "rdp_url": None,
                            "data_license_attributes": {
                                "non_commercial": False,
                                "attribution": True,
                                "share_alike": False
                            }
                        }
                    },
                    {
                        "source": "NCIt",
                        "match_type": 80,
                        "records": [
                            {
                                "label": "Cisplatin",
                                "concept_id": "ncit:C376",
                                "aliases": [],
                                "xrefs": ["chemidplus:15663-27-1"],
                                "associated_with": [
                                    "umls:C0008838",
                                    "fda:Q20Q21Q62J",
                                    "chebi:CHEBI:27899"
                                ],
                                "approval_ratings": None,
                                "trade_names": []
                            }
                        ],
                        "source_meta_": {
                            "data_license": "CC BY 4.0",
                            "data_license_url": "https://creativecommons.org/licenses/by/4.0/legalcode",  # noqa: E501
                            "version": "20.09d",
                            "data_url": "https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/archive/2020/20.09d_Release/",  # noqa: E501
                            "rdp_url": "http://reusabledata.org/ncit.html",
                            "data_license_attributes": {
                                "non_commercial": False,
                                "attribution": True,
                                "share_alike": False
                            }
                        }
                    },
                    {
                        "source": "Wikidata",
                        "match_type": 80,
                        "records": [
                            {
                                "label": "cisplatin",
                                "concept_id": "wikidata:Q412415",
                                "aliases": [
                                    "Platinol",
                                    "cis-diamminedichloroplatinum(II)",
                                    "CDDP",
                                    "Cis-DDP",
                                    "CIS-DDP",
                                    "Platinol-AQ"
                                ],
                                "xrefs": [
                                    "chemidplus:15663-27-1",
                                    "chembl:CHEMBL11359",
                                    "rxcui:2555",
                                    "drugbank:DB00515"
                                ],
                                "associated_with": [
                                    "pubchem.compound:5702198"
                                ],
                                "approval_ratings": None,
                                "trade_names": []
                            }
                        ],
                        "source_meta_": {
                            "data_license": "CC0 1.0",
                            "data_license_url": "https://creativecommons.org/publicdomain/zero/1.0/",  # noqa: E501
                            "version": "20210331",
                            "data_url": None,
                            "rdp_url": None,
                            "data_license_attributes": {
                                "non_commercial": False,
                                "attribution": False,
                                "share_alike": False
                            }
                        }
                    }
                ],
                "service_meta_": {
                    "name": "thera-py",
                    "version": "0.1.0",
                    "response_datetime": "2021-04-05T16:44:15.367831",
                    "url": "https://github.com/cancervariants/therapy-normalization"
                }
            }



"""ETL methods for NCIt source"""
import logging
from typing import Set

import requests
import owlready2 as owl
from owlready2.entity import ThingClass

from therapy import DownloadException
from therapy.schemas import SourceName, NamespacePrefix, SourceMeta, RecordParams
from therapy.etl.base import Base

logger = logging.getLogger("therapy")
logger.setLevel(logging.DEBUG)


class NCIt(Base):
    """Class for NCIt ETL methods.

    Extracts:
     * NCIt classes with semantic_type "Pharmacologic Substance" but not
       Retired_Concept
     * NCIt classes that are subclasses of C1909 (Pharmacologic Substance)
    """

    def _download_data(self) -> None:
        """Download NCI thesaurus source file.
        The NCI directory structure can be a little tricky, so this method attempts to
        retrieve a file matching the latest version number from both the subdirectory
        root (where the current version is typically posted) as well as the year-by-year
        archives if that fails.
        """
        logger.info("Retrieving source data for NCIt")
        base_url = "https://evs.nci.nih.gov/ftp1/NCI_Thesaurus"
        # ping base NCIt directory
        release_fname = f"Thesaurus_{self._version}.OWL.zip"
        src_url = f"{base_url}/{release_fname}"
        r_try = requests.get(src_url)
        if r_try.status_code != 200:
            # ping NCIt archive directories
            archive_url = f"{base_url}/archive/{self._version}_Release/{release_fname}"
            archive_try = requests.get(archive_url)
            if archive_try.status_code != 200:
                old_archive_url = f"{base_url}/archive/20{self._version[0:2]}/{self._version}_Release/{release_fname}"  # noqa: E501
                old_archive_try = requests.get(old_archive_url)
                if old_archive_try.status_code != 200:
                    msg = (
                        f"NCIt download failed: tried {src_url}, {archive_url}, and "
                        f"{old_archive_url}"
                    )
                    logger.error(msg)
                    raise DownloadException(msg)
                else:
                    src_url = old_archive_url
            else:
                src_url = archive_url

        self._http_download(src_url, self._src_dir / f"ncit_{self._version}.owl",
                            handler=self._zip_handler)
        logger.info("Successfully retrieved source data for NCIt")

    def _get_desc_nodes(self, node: ThingClass,
                        uq_nodes: Set[ThingClass]) -> Set[ThingClass]:
        """Create set of unique subclasses of node parameter.
        Should be originally called on ncit:C1909: Pharmacologic Substance.

        :param owlready2.entity.ThingClass node: concept node to either retrieve
            descendants of, or to normalize and add to DB
        :param Set[owlready2.entity.ThingClass] uq_nodes: set of unique class nodes
            found so far from recursive tree exploration
        :return: the uq_nodes set, updated with any class nodes found from recursive
            exploration of this branch of the class tree
        :rtype: Set[owlready2.entity.ThingClass]
        """
        children = node.descendants()
        if children:
            for child_node in children:
                if child_node is not node:
                    uq_nodes.add(child_node)
                    self._get_desc_nodes(child_node, uq_nodes)
        return uq_nodes

    def _get_typed_nodes(self, uq_nodes: Set[ThingClass],
                         ncit: owl.namespace.Ontology) -> Set[ThingClass]:
        """Get all nodes with semantic_type Pharmacologic Substance

        :param Set[owlready2.entity.ThingClass] uq_nodes: set of unique class nodes
            found so far.
        :param owl.namespace.Ontology ncit: owlready2 Ontology instance for NCIt.
        :return: uq_nodes, with the addition of all classes found to have semantic_type
            Pharmacologic Substance and not of type Retired_Concept
        :rtype: Set[owlready2.entity.ThingClass]
        """
        graph = owl.default_world.as_rdflib_graph()

        query_str = """SELECT ?x WHERE {
            ?x <http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#P106>
            \"Pharmacologic Substance\"
        }
        """
        typed_results = set(graph.query(query_str))

        retired_query_str = """SELECT ?x WHERE {
            ?x <http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#P310>
            \"Retired_Concept\"
        }
        """
        retired_results = set(graph.query(retired_query_str))

        typed_results = {r for r in (typed_results - retired_results)
                         if r is not None}

        for result in typed_results:
            # parse result as URI and get ThingClass object back from NCIt
            class_object = ncit[result[0].toPython().split("#")[1]]
            uq_nodes.add(class_object)
        return uq_nodes

    def _transform_data(self) -> None:
        """Get data from file and construct objects for loading"""
        ncit = owl.get_ontology(self._src_file.absolute().as_uri())
        ncit.load()
        uq_nodes = {ncit.C49236}  # add Therapeutic Procedure
        uq_nodes = self._get_desc_nodes(ncit.C1909, uq_nodes)
        uq_nodes = self._get_typed_nodes(uq_nodes, ncit)
        for node in uq_nodes:
            concept_id = f"{NamespacePrefix.NCIT.value}:{node.name}"
            if node.P108:
                label = node.P108.first()
            else:
                label = None
            aliases = node.P90.copy()  # prevent CallbackList error
            if label and aliases and label in aliases:
                aliases.remove(label)

            xrefs = []
            associated_with = []
            if node.P207:
                associated_with.append(f"{NamespacePrefix.UMLS.value}:"
                                       f"{node.P207.first()}")
            if node.P210:
                xrefs.append(f"{NamespacePrefix.CASREGISTRY.value}:"
                             f"{node.P210.first()}")
            if node.P319:
                associated_with.append(f"{NamespacePrefix.UNII.value}:"
                                       f"{node.P319.first()}")
            if node.P368:
                iri = node.P368.first()
                if ":" in iri:
                    iri = iri.split(":")[1]
                associated_with.append(f"{NamespacePrefix.CHEBI.value}:{iri}")
            params: RecordParams = {  # type: ignore
                "concept_id": concept_id,
                "label": label,
                "aliases": aliases,
                "xrefs": xrefs,
                "associated_with": associated_with
            }
            self._load_therapy(params)

    def _load_meta(self) -> None:
        """Load metadata"""
        metadata = SourceMeta(
            data_license="CC BY 4.0",
            data_license_url="https://creativecommons.org/licenses/by/4.0/legalcode",  # noqa F401
            version=self._version,
            data_url="https://evs.nci.nih.gov/ftp1/NCI_Thesaurus/",
            rdp_url="http://reusabledata.org/ncit.html",
            data_license_attributes={
                "non_commercial": False,
                "share_alike": False,
                "attribution": True
            }
        )
        params = dict(metadata)
        params["src_name"] = SourceName.NCIT.value
        self.database.metadata.put_item(Item=params)



"""Module for Guide to PHARMACOLOGY ETL methods."""
from typing import Optional, Dict, Any, List, Union
import csv
import html
from pathlib import Path
import re

import requests

from therapy import logger
from therapy.schemas import SourceMeta, SourceName, NamespacePrefix, ApprovalRating
from therapy.etl.base import Base, SourceFormatException


TAG_PATTERN = re.compile("</?[a-zA-Z]+>")
PMID_PATTERN = re.compile(r"\[PMID:[ ]?\d+\]")


class GuideToPHARMACOLOGY(Base):
    """Class for Guide to PHARMACOLOGY ETL methods."""

    def _download_data(self) -> None:
        """Download the latest version of Guide to PHARMACOLOGY."""
        logger.info("Retrieving source data for Guide to PHARMACOLOGY")
        if not self._ligands_file.exists():
            self._http_download("https://www.guidetopharmacology.org/DATA/ligands.tsv",
                                self._ligands_file)
            assert self._ligands_file.exists()
        if not self._mapping_file.exists():
            self._http_download("https://www.guidetopharmacology.org/DATA/ligand_id_mapping.tsv",  # noqa: E501
                                self._mapping_file)
            assert self._mapping_file.exists()
        logger.info("Successfully retrieved source data for Guide to PHARMACOLOGY")

    def _download_file(self, file_url: str, fn: str) -> None:
        """Download individual data file.

        :param str file_url: Data url for file
        :param str fn: File name
        """
        r = requests.get(file_url)
        if r.status_code == 200:
            prefix = SourceName.GUIDETOPHARMACOLOGY.value.lower()
            path = self._src_dir / f"{prefix}_{fn}_{self._version}.tsv"
            if fn == "ligands":
                self._ligands_file: Path = path
            else:
                self._mapping_file: Path = path
            with open(str(path), "wb") as f:
                f.write(r.content)

    def _extract_data(self, use_existing: bool = False) -> None:
        """Gather GtoPdb source files.

        :param bool use_existing: if True, don't try to fetch latest source data
        """
        self._src_dir.mkdir(exist_ok=True, parents=True)
        prefix = SourceName.GUIDETOPHARMACOLOGY.value.lower()
        if use_existing:
            ligands_files = list(sorted(self._src_dir.glob(f"{prefix}_ligands_*.tsv")))
            if len(ligands_files) < 1:
                raise FileNotFoundError("No GtoPdb ligands files found")

            for ligands_file in ligands_files[::-1]:
                try:
                    version = self._parse_version(
                        ligands_file,
                        re.compile(prefix + r"_ligands_(.+)\.tsv")
                    )
                except FileNotFoundError:
                    raise FileNotFoundError(
                        "Unable to parse GtoPdb version value from ligands file "
                        f"located at {ligands_file.absolute().as_uri()} -- check "
                        "filename against schema defined in README: "
                        "https://github.com/cancervariants/therapy-normalization#update-sources"  # noqa: E501
                    )
                check_mapping_file = self._src_dir / f"{prefix}_ligand_id_mapping_{version}.tsv"  # noqa: E501
                if check_mapping_file.exists():
                    self._version = version
                    self._ligands_file = ligands_file
                    self._mapping_file = check_mapping_file
                    break
            if self._mapping_file is None:
                raise FileNotFoundError(
                    "Unable to find complete GtoPdb data set with matching version "
                    "values. Check filenames against schema defined in README: "
                    "https://github.com/cancervariants/therapy-normalization#update-sources"  # noqa: E501
                )
        else:
            self._version = self.get_latest_version()
            self._ligands_file = self._src_dir / f"{prefix}_ligands_{self._version}.tsv"
            self._mapping_file = self._src_dir / f"{prefix}_ligand_id_mapping_{self._version}.tsv"  # noqa: E501
            if not (self._ligands_file.exists() and self._mapping_file.exists()):
                self._download_data()
                assert self._ligands_file.exists()
                assert self._mapping_file.exists()

    def _transform_data(self) -> None:
        """Transform Guide To PHARMACOLOGY data."""
        data: Dict[str, Any] = dict()
        self._transform_ligands(data)
        self._transform_ligand_id_mappings(data)
        for param in data.values():
            self._load_therapy(param)

    @staticmethod
    def _process_name(name: str) -> str:
        """Remove tags from GtoP name object.
        :param name: raw drug referent
        :return: cleaned name (may be unchanged)
        """
        name = re.sub(TAG_PATTERN, "", name)
        return name

    def _transform_ligands(self, data: Dict) -> None:
        """Transform ligands data file and add this data to `data`.

        :param dict data: Transformed data
        """
        with open(self._ligands_file, "r") as f:
            rows = csv.reader(f, delimiter="\t")

            # check that file structure is the same
            next(rows)
            if next(rows) != [
                "Ligand ID", "Name", "Species", "Type", "Approved", "Withdrawn",
                "Labelled", "Radioactive", "PubChem SID", "PubChem CID", "UniProt ID",
                "Ensembl ID", "Ligand Subunit IDs", "Ligand Subunit Name",
                "Ligand Subunit UniProt IDs", "Ligand Subunit Ensembl IDs",
                "IUPAC name", "INN", "Synonyms", "SMILES", "InChIKey", "InChI",
                "GtoImmuPdb", "GtoMPdb", "Antibacterial"
            ]:
                raise SourceFormatException(
                    "GtoP ligands file contains missing or unrecognized columns. See "
                    "FAQ in README for suggested resolution."
                )

            for row in rows:
                params: Dict[str, Union[List[str], str]] = {
                    "concept_id":
                        f"{NamespacePrefix.GUIDETOPHARMACOLOGY.value}:{row[0]}",
                    "label": self._process_name(row[1]),
                    "src_name": SourceName.GUIDETOPHARMACOLOGY.value
                }

                approval_rating = self._set_approval_rating(row[4], row[5])
                if approval_rating:
                    params["approval_ratings"] = [approval_rating]

                associated_with = list()
                aliases = list()
                if row[8]:
                    associated_with.append(f"{NamespacePrefix.PUBCHEMSUBSTANCE.value}:{row[8]}")  # noqa: E501
                if row[9]:
                    associated_with.append(f"{NamespacePrefix.PUBCHEMCOMPOUND.value}:{row[9]}")  # noqa: E501
                if row[10]:
                    associated_with.append(f"{NamespacePrefix.UNIPROT.value}:{row[10]}")
                if row[16]:
                    aliases.append(self._process_name(row[16]))  # IUPAC
                if row[17]:
                    # International Non-proprietary Name assigned by the WHO
                    aliases.append(self._process_name(row[17]))
                if row[18]:
                    # synonyms
                    synonyms = row[18].split("|")
                    for s in synonyms:
                        if "&" in s and ";" in s:
                            name_code = s[s.index("&"):s.index(";") + 1]
                            if name_code.lower() in ["&reg;", "&trade;"]:
                                # Remove trademark symbols to allow for search
                                s = s.replace(name_code, "")
                            s = html.unescape(s)
                        aliases.append(self._process_name(s))
                if row[20]:
                    associated_with.append(f"{NamespacePrefix.INCHIKEY.value}:{row[20]}")  # noqa: E501

                if associated_with:
                    params["associated_with"] = associated_with
                if aliases:
                    params["aliases"] = aliases

                data[params["concept_id"]] = params

    @staticmethod
    def _get_xrefs(ref: str, namespace: str) -> List[str]:
        """Construct xrefs from raw string.
        :param ref: raw ref value (may need to be separated)
        :param namspace: namespace prefix to use
        :return: List (usually with just one member) of xref IDs
        """
        xrefs = []
        for split_ref in ref.split("|"):
            xref = f"{namespace}:{split_ref}"
            xrefs.append(xref)
        return xrefs

    def _transform_ligand_id_mappings(self, data: Dict) -> None:
        """Transform ligand_id_mappings and add this data to `data`
        All ligands found in this file should already be in data

        :param dict data: Transformed data
        """
        with open(self._mapping_file.absolute(), "r") as f:
            rows = csv.reader(f, delimiter="\t")
            next(rows)
            if next(rows) != [
                "Ligand id", "Name", "Species", "Type", "PubChem SID", "PubChem CID",
                "ChEMBl ID", "Chebi ID", "UniProt id", "Ensembl ID", "IUPAC name",
                "INN", "CAS", "DrugBank ID", "Drug Central ID"
            ]:
                raise SourceFormatException(
                    "GtoP ligand mapping file contains missing or unrecognized "
                    "columns. See FAQ in README for suggested resolution."
                )

            for row in rows:
                concept_id = f"{NamespacePrefix.GUIDETOPHARMACOLOGY.value}:{row[0]}"

                if concept_id not in data:
                    logger.debug(f"{concept_id} not in ligands")
                    continue
                params = data[concept_id]
                xrefs = list()
                associated_with = params.get("associated_with", [])
                if row[6]:
                    xrefs += self._get_xrefs(row[6], NamespacePrefix.CHEMBL.value)
                if row[7]:
                    # CHEBI IDs are already namespaced
                    associated_with.append(row[7])
                if row[8]:
                    associated_with += self._get_xrefs(
                        row[8], NamespacePrefix.UNIPROT.value
                    )
                if row[12]:
                    xrefs += self._get_xrefs(row[12], NamespacePrefix.CASREGISTRY.value)
                if row[13]:
                    xrefs += self._get_xrefs(row[13], NamespacePrefix.DRUGBANK.value)
                if row[14]:
                    associated_with += self._get_xrefs(
                        row[14], NamespacePrefix.DRUGCENTRAL.value
                    )
                if xrefs:
                    params["xrefs"] = xrefs
                if associated_with:
                    params["associated_with"] = associated_with

    def _set_approval_rating(self, approved: str,
                             withdrawn: str) -> Optional[str]:
        """Set approval rating value.

        :param str approved: The drug is or has in the past been approved for human
            clinical use by a regulatory agency
        :param str withdrawn: The drug is no longer approved for its original clinical
            use in one or more countries
        :return: Approval rating
        """
        if approved and not withdrawn:
            approval_rating: Optional[str] = ApprovalRating.GTOPDB_APPROVED.value
        elif withdrawn:
            approval_rating = ApprovalRating.GTOPDB_WITHDRAWN.value
        else:
            approval_rating = None
        return approval_rating

    def _load_meta(self) -> None:
        """Load Guide to PHARMACOLOGY metadata to database."""
        meta = SourceMeta(
            data_license="CC BY-SA 4.0",
            data_license_url="https://creativecommons.org/licenses/by-sa/4.0/",
            version=self._version,
            data_url="https://www.guidetopharmacology.org/download.jsp",
            rdp_url=None,
            data_license_attributes={
                "non_commercial": False,
                "share_alike": True,
                "attribution": True,
            }
        )
        params = dict(meta)
        params["src_name"] = SourceName.GUIDETOPHARMACOLOGY.value
        self.database.metadata.put_item(Item=params)



"""Create concept groups and merged records."""
from typing import Set, Dict, Optional, Any, Tuple
import logging
from timeit import default_timer as timer

from therapy.schemas import SourcePriority
from therapy.database import Database


logger = logging.getLogger("therapy")
logger.setLevel(logging.DEBUG)


class Merge:
    """Handles record merging."""

    def __init__(self, database: Database) -> None:
        """Initialize Merge instance.

        * self._groups is a dictionary keying concept IDs to the Set of concept IDs in
        that group. This is redundantly captured for all group members (i.e., a group
        of 5 concepts would all have their own key, and the value would be an identical
        Set of the same IDs)
        * self._unii_to_drugsatfda keys UNII codes to valid Drugs@FDA concepts.
        Because UNIIs aren't stored directly in groups, we utilize a separate lookup
        table to prevent repeat queries.
        * self._failed_lookups stores concept IDs for which lookups have been attempted
        and failed. Because we don't associate these IDs with groups, a separate
        mapping is necessary to prevent repeat queries.

        :param Database database: db instance to use for record retrieval and creation.
        """
        self.database = database
        self._groups: Dict[str, Set[str]] = {}
        self._unii_to_drugsatfda: Dict[str, Set[str]] = {}
        self._failed_lookups: Set[str] = set()

    def create_merged_concepts(self, record_ids: Set[str]) -> None:
        """Create concept groups, generate merged concept records, and update database.

        :param Set[str] record_ids: concept identifiers from which groups should be
            generated.
        """
        self._create_record_id_sets(record_ids)

        # don't create separate records for single-member groups
        self._groups = {k: v for k, v in self._groups.items() if len(v) > 1}

        logger.info("Creating merged records and updating database...")
        uploaded_ids = set()
        start = timer()
        for record_id, group in self._groups.items():
            if record_id in uploaded_ids:
                continue
            merged_record = self._generate_merged_record(group)

            # add group merger item to DB
            self.database.add_record(merged_record, "merger")

            # add updated references
            for concept_id in group:
                if not self.database.get_record_by_id(concept_id, False):
                    logger.error(f"Updating nonexistent record: {concept_id} "
                                 f"for {merged_record['label_and_type']}")
                else:
                    merge_ref = merged_record["concept_id"].lower()
                    self.database.update_record(concept_id, "merge_ref", merge_ref)
            uploaded_ids |= group
        logger.info("Merged concept generation successful.")
        end = timer()
        logger.debug(f"Generated and added concepts in {end - start} seconds")

    def _rxnorm_brand_lookup(self, record_id: str) -> Optional[str]:
        """Rx Norm provides brand references back to original therapy concepts.  This
        routine checks whether an ID from a concept group requires an additional
        dereference to obtain the original RxNorm record.

        :param str record_id: RxNorm concept ID to check
        :return: concept ID for RxNorm record if successful, None otherwise
        """
        brand_lookup = self.database.get_records_by_type(record_id, "rx_brand")
        n = len(brand_lookup)
        if n == 1:
            lookup_id = brand_lookup[0]["concept_id"]
            db_record = self.database.get_record_by_id(lookup_id,
                                                       False)
            if db_record:
                return db_record["concept_id"]
        elif n > 1:
            logger.warning(f"Brand lookup for {record_id} had {n} matches")
        return None

    def _get_drugsatfda_from_unii(self, ref: Dict) -> Optional[str]:
        """Given an `associated_with` item keying a UNII code to a Drugs@FDA record,
        verify that the record can be safely added to a concept group.
        Drugs@FDA tracks a number of "compound therapies", and provides UNIIs to each
        individual component. If we included them in normalized record sets, they would
        end up merging distinct therapies under the umbrella of the compound group.
        We're excluding Drugs@FDA records with multiple UNIIs as a tentative solution.

        :param Dict ref: `associated_with` item where `label_and_type` includes
            a UNII code and `src_name` == "DrugsAtFDA"
        :return: Drugs@FDA concept ID if record meets rules, None otherwise
        """
        concept_id = ref["concept_id"]
        fetched = self.database.get_record_by_id(concept_id, False)
        if fetched:
            uniis = [a for a
                     in fetched.get("associated_with", [])
                     if a.startswith("unii")]
        else:
            logger.error(f"Couldn't retrieve record for {concept_id} from {ref}")
            return None
        if len(uniis) == 1:
            return fetched["concept_id"]
        else:
            return None

    def _get_xrefs(self, record: Dict[str, Any]) -> Set[str]:
        """Extract references to entries in other sources from a record.

        :param Dict record: record to process
        :return: Set of xref values
        """
        xrefs = set(record.get("xrefs", []))
        unii_xrefs = [a for a in record.get("associated_with", [])
                      if a.startswith("unii:")]
        for unii in unii_xrefs:
            # get Drugs@FDA records that are associated_with this UNII
            drugsatfda_ids = self._unii_to_drugsatfda.get(unii)
            if drugsatfda_ids is not None:
                xrefs |= drugsatfda_ids
                continue

            unii_assoc = self.database.get_records_by_type(unii.lower(),
                                                           "associated_with")
            drugsatfda_refs = set()
            for ref in unii_assoc:
                if ref["src_name"] == "DrugsAtFDA":
                    drugsatfda_ref = self._get_drugsatfda_from_unii(ref)
                    if drugsatfda_ref:
                        drugsatfda_refs.add(drugsatfda_ref)
            self._unii_to_drugsatfda[unii] = drugsatfda_refs
            xrefs |= drugsatfda_refs
        return xrefs

    def _create_record_id_set(
        self, record_id: str, observed_id_set: Optional[Set] = None
    ) -> Set[str]:
        """Create concept ID group for an individual record ID.

        :param str record_id: concept ID for record to build group from
        :return: set of related identifiers pertaining to a common concept.
        """
        if observed_id_set is None:
            observed_id_set = set()

        if record_id in self._groups:
            return self._groups[record_id]
        elif record_id.startswith("drugsatfda"):
            return {record_id}
        elif record_id in self._failed_lookups:
            return observed_id_set - {record_id}

        # get record
        db_record = self.database.get_record_by_id(record_id)
        if not db_record:
            if record_id.startswith("rxcui"):
                brand_lookup = self._rxnorm_brand_lookup(record_id)
                if brand_lookup:
                    return self._create_record_id_set(brand_lookup, observed_id_set)
            logger.warning(f"Unable to resolve lookup for {record_id} in "
                           f"ID set: {observed_id_set}")
            self._failed_lookups.add(record_id)
            return observed_id_set - {record_id}

        # construct group
        local_id_set = self._get_xrefs(db_record)
        if not local_id_set:
            return observed_id_set | {db_record["concept_id"]}
        merged_id_set = {record_id} | observed_id_set
        for local_record_id in local_id_set - observed_id_set:
            merged_id_set |= self._create_record_id_set(local_record_id, merged_id_set)
        return merged_id_set

    def _create_record_id_sets(self, record_ids: Set[str]) -> None:
        """Update self._groups with normalized concept groups.
        :param Set[str] record_ids: concept identifiers from which groups should be
            generated.
        """
        logger.info("Generating record ID sets...")
        start = timer()
        for record_id in record_ids:
            new_group = self._create_record_id_set(record_id)
            if new_group:
                for concept_id in new_group:
                    self._groups[concept_id] = new_group
        end = timer()
        logger.debug(f"Built record ID sets in {end - start} seconds")

    def _generate_merged_record(self, record_id_set: Set[str]) -> Dict:
        """Generate merged record from provided concept ID group.
        Where attributes are 'set-like', they should be combined, and where
        they are 'scalar-like', assign from the highest-priority source where
        that attribute is not None.

        Uses the SourcePriority schema to define source priority.

        :param Set record_id_set: group of concept IDs
        :return: completed merged drug object to be stored in DB
        """
        records = []
        for record_id in record_id_set:
            record = self.database.get_record_by_id(record_id)
            if record:
                records.append(record)
            else:
                logger.error(f"Merge record generator could not retrieve "
                             f"record for {record_id} in {record_id_set}")

        def record_order(record: Dict) -> Tuple[int, str]:
            """Provide priority values of concepts for sort function."""
            src = record["src_name"].upper()
            if src == "DRUGS@FDA":
                src = "DRUGSATFDA"
            if src in SourcePriority.__members__:
                source_rank = SourcePriority[src].value
            else:
                raise Exception(f"Prohibited source: {src} in concept_id "
                                f"{record['concept_id']}")
            return source_rank, record["concept_id"]
        records.sort(key=record_order)

        # initialize merged record
        merged_attrs = {
            "concept_id": records[0]["concept_id"],
            "xrefs": [r["concept_id"] for r in records[1:]],
            "label": None,
            "aliases": set(),
            "trade_names": set(),
            "associated_with": set(),
            "approval_ratings": set(),
            "approval_year": set(),
            "has_indication": [],
        }

        # merge from constituent records
        set_fields = ["aliases", "trade_names", "associated_with", "approval_year"]
        for record in records:
            for field in set_fields:
                merged_attrs[field] |= set(record.get(field, set()))
            approval_ratings = record.get("approval_ratings")
            if approval_ratings:
                merged_attrs["approval_ratings"] |= set(approval_ratings)
            if merged_attrs["label"] is None:
                merged_attrs["label"] = record.get("label")
            for ind in record.get("has_indication", []):
                if ind not in merged_attrs["has_indication"]:
                    merged_attrs["has_indication"].append(ind)

        # clear unused fields
        for field in set_fields + ["has_indication", "approval_ratings"]:
            field_value = merged_attrs[field]
            if field_value:
                merged_attrs[field] = list(field_value)
            else:
                del merged_attrs[field]
        if not merged_attrs["label"]:
            del merged_attrs["label"]

        merged_attrs["label_and_type"] = f"{merged_attrs['concept_id'].lower()}##merger"
        merged_attrs["item_type"] = "merger"
        return merged_attrs
